import os
from unittest import TestCase
from unittest.mock import MagicMock

import pytest

from recce.adapter.dbt_adapter import load_manifest, load_catalog, DbtAdapter
from recce.core import RecceContext, set_default_context
from recce.run import schema_diff_should_be_approved


def test_validator():
    from recce.tasks.schema import SchemaDiffCheckValidator
    validator = SchemaDiffCheckValidator()

    def validate(params: dict):
        validator.validate({
            'name': 'test',
            'type': 'schema_diff',
            'params': params,
        })

    # Select all models
    validate({})

    # Select by node name
    validate({
        'node_id': 'abc',
    })
    validate({
        'node_id': ['abc'],
    })

    # Select by selector
    validate({
        'select': 'customers',
        'exclude': 'customers',
        'packages': ['jaffle_shop'],
        'view_mode': 'all',
    })

    # packages should be an array
    with pytest.raises(ValueError):
        validate({
            'packages': 'jaffle_shop',
        })

    # view_mode should be 'all' or 'changed_models'
    validate({
        'view_mode': None,
    })
    validate({
        'view_mode': 'all',
    })
    with pytest.raises(ValueError):
        validate({
            'view_mode': 'abc',
        })


test_root_path = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


class TestSchemaDiffAutoApprove(TestCase):

    def setUp(self):
        self.default_context = MagicMock(spec=RecceContext)
        manifest = load_manifest(path=os.path.join(test_root_path, 'manifest.json'))
        catalog = load_catalog(path=os.path.join(test_root_path, 'catalog.json'))
        dbt_adapter = DbtAdapter(curr_manifest=manifest, curr_catalog=catalog)
        self.default_context.adapter = dbt_adapter

        dbt_adapter.select_nodes = MagicMock()
        # Base and Current will be the same
        self.default_context.get_lineage.return_value = dbt_adapter.get_lineage()
        set_default_context(self.default_context)

    def test_schema_diff_should_be_approved(self):
        # Node_id is string
        is_approved = schema_diff_should_be_approved({
            'node_id': 'model.jaffle_shop.customers',
        })
        assert is_approved is True

        # Node_id is list
        is_approved = schema_diff_should_be_approved({
            'node_id': ['model.jaffle_shop.customers'],
        })
        assert is_approved is True

        # Select all models
        self.default_context.adapter.select_nodes.return_value = ['model.jaffle_shop.customers']
        is_approved = schema_diff_should_be_approved({
            'select': 'customers',
        })
        assert is_approved is True



import os

from git import Repo, InvalidGitRepositoryError


def current_default_branch():
    try:
        repo = Repo(search_parent_directories=True)
        return repo.remotes.origin.refs['HEAD'].reference.remote_head
    except Exception:
        return None


def current_branch():
    try:
        repo = Repo(search_parent_directories=True)
        if not repo.active_branch:
            return None
        return repo.active_branch.name
    except TypeError:
        # happened when HEAD is a detached symbolic reference
        return None
    except Exception:
        return None


def current_commit_hash(short_length: int = 7, short: bool = False):
    try:
        repo = Repo(search_parent_directories=True)
        if not repo.active_branch:
            return None
        if short:
            return repo.head.commit.hexsha[:short_length]
        return repo.head.commit.hexsha
    except TypeError:
        # happened when HEAD is a detached symbolic reference
        return None
    except Exception:
        return None


def commit_hash_from_branch(branch: str, short_length: int = 7, short: bool = False):
    try:
        repo = Repo(search_parent_directories=True)
        commit = repo.commit(branch)
        if short:
            return commit.hexsha[:short_length]
        return commit.hexsha
    except TypeError:
        # happened when HEAD is a detached symbolic reference
        return None
    except Exception:
        return None


def hosting_repo(remote: str = 'origin'):
    try:
        repo = Repo(search_parent_directories=True)
        origin_url = repo.remote(name=remote).url
        remote_repo = None

        if origin_url.startswith('git@'):
            # Handle git@github.com:user/repo.git
            remote_repo = origin_url.split(':')[1].replace('.git', '')

        elif origin_url.startswith('https://') or origin_url.startswith('http://'):
            # Handle https://github.com/user/repo.git or http://github.com/user/repo.git
            remote_repo = '/'.join(origin_url.split('/')[-2:]).replace('.git', '')

        return remote_repo
    except ValueError:
        repo = Repo(search_parent_directories=True)
        toplevel_dir = repo.git.rev_parse("--show-toplevel")

        return os.path.basename(toplevel_dir)
    except InvalidGitRepositoryError:
        return None



import pytest

from recce.tasks import ProfileDiffTask


def test_profile_diff(dbt_test_helper):
    csv_data_curr = """
        customer_id,name,age
        1,Alice,30
        2,Bob,25
        3,Charlie,35
        """

    csv_data_base = """
        customer_id,name,age
        1,Alice,35
        2,Bob,25
        3,Charlie,35
        """

    dbt_test_helper.create_model("customers", csv_data_base, csv_data_curr)
    params = dict(model='customers')
    task = ProfileDiffTask(params)
    run_result = task.execute()

    assert len(run_result.current.data) == 3
    assert len(run_result.base.data) == 3


def test_validator():
    from recce.tasks.profile import ProfileDiffCheckValidator

    def validate(params: dict = {}, view_options: dict = {}):
        ProfileDiffCheckValidator().validate({
            'name': 'test',
            'type': 'profile_diff',
            'params': params,
            'view_options': view_options,
        })

    validate({
        'model': 'customers',
    })

    with pytest.raises(ValueError):
        validate({
        })



import textwrap
from typing import List

from pydantic import BaseModel

from .core import Task, TaskResultDiffer, CheckValidator
from .dataframe import DataFrame
from ..core import default_context
from ..exceptions import RecceException
from ..models import Check


class ProfileDiffParams(BaseModel):
    model: str


class ProfileResult(BaseModel):
    base: DataFrame
    current: DataFrame


class ProfileDiffTask(Task):

    def __init__(self, params):
        super().__init__()
        self.params = ProfileDiffParams(**params)
        self.connection = None

    def execute(self):
        import agate
        from recce.adapter.dbt_adapter import DbtAdapter, merge_tables
        dbt_adapter: DbtAdapter = default_context().adapter

        model: str = self.params.model

        self._verify_dbt_profiler(dbt_adapter)

        with dbt_adapter.connection_named("query"):
            self.connection = dbt_adapter.get_thread_connection()

            base_columns = [column for column in dbt_adapter.get_columns(model, base=True)]
            curr_columns = [column for column in dbt_adapter.get_columns(model, base=False)]
            total = len(base_columns) + len(curr_columns)
            completed = 0

            tables: List[agate.Table] = []

            for column in base_columns:
                self.update_progress(message=f'[Base] Profile column: {column.name}', percentage=completed / total)
                relation = dbt_adapter.create_relation(model, base=True)
                response, table = self._profile_column(dbt_adapter, relation, column)
                tables.append(table)
                completed = completed + 1
                self.check_cancel()
            base = DataFrame.from_agate(merge_tables(tables))

            tables: List[agate.Table] = []
            for column in curr_columns:
                self.update_progress(message=f'[Current] Profile column: {column.column}', percentage=completed / total)
                relation = dbt_adapter.create_relation(model, base=False)
                response, table = self._profile_column(dbt_adapter, relation, column)
                tables.append(table)
                completed = completed + 1
                self.check_cancel()
            current = DataFrame.from_agate(merge_tables(tables))

            return ProfileResult(base=base, current=current)

    def _verify_dbt_profiler(self, dbt_adapter):
        for macro_name, macro in dbt_adapter.manifest.macros.items():
            if macro.package_name == 'dbt_profiler':
                break
        else:
            raise RecceException(
                r"Package 'dbt_profiler' not found. Please refer to the link to install: https://hub.getdbt.com/data-mie/dbt_profiler/")

    def _profile_column(self, dbt_adapter, relation, column):
        sql_template = textwrap.dedent(r"""
        select
        '{{column_name}}' as column_name,
        nullif('{{column_type}}', '') as data_type,
        {{ dbt_profiler.measure_row_count(column_name, column_type) }} as row_count,
        {{ dbt_profiler.measure_not_null_proportion(column_name, column_type) }} as not_null_proportion,
        {{ dbt_profiler.measure_distinct_proportion(column_name, column_type) }} as distinct_proportion,
        {{ dbt_profiler.measure_distinct_count(column_name, column_type) }} as distinct_count,
        {{ dbt_profiler.measure_is_unique(column_name, column_type) }} as is_unique,
        {{ dbt_profiler.measure_min(column_name, column_type) }} as min,
        {{ dbt_profiler.measure_max(column_name, column_type) }} as max,
        {{ dbt_profiler.measure_avg(column_name, column_type) }} as avg,
        {{ dbt_profiler.measure_median(column_name, column_type) }} as median
        from
        {{ relation }}
        """)
        column_name = column.name
        column_type = column.data_type.lower()
        db_type = dbt_adapter.adapter.type()
        if db_type == 'bigquery' and column_type.startswith('array'):
            # DRC-663: Support bigquery array type
            sql_template = textwrap.dedent(r"""
            select
            '{{column_name}}' as column_name,
            nullif('{{column_type}}', '') as data_type,
            {{ dbt_profiler.measure_row_count(column_name, column_type) }} as row_count,
            {{ dbt_profiler.measure_not_null_proportion(column_name, column_type) }} as not_null_proportion,
            cast(null as {{ dbt.type_numeric() }}) as distinct_proportion,
            cast(null as {{ dbt.type_numeric() }}) as distinct_count,
            null as is_unique,
            cast(min(ARRAY_LENGTH({{ adapter.quote(column_name) }})) as {{ dbt_profiler.type_string() }}) as min,
            cast(max(ARRAY_LENGTH({{ adapter.quote(column_name) }})) as {{ dbt_profiler.type_string() }}) as max,
            avg(ARRAY_LENGTH({{ adapter.quote(column_name) }})) as avg,
            APPROX_QUANTILES(ARRAY_LENGTH({{ adapter.quote(column_name) }}), 100)[OFFSET(50)] as median,
            from
            {{ relation }}
            """)

        try:
            sql = dbt_adapter.generate_sql(
                sql_template,
                base=False,  # always false because we use the macro in current manifest
                context=dict(relation=relation, column_name=column_name, column_type=column_type)
            )
        except Exception as e:
            raise RecceException(f"Failed to generate SQL for profiling column: {column_name}") from e

        try:
            return dbt_adapter.execute(sql, fetch=True)
        except Exception as e:
            from recce.adapter.dbt_adapter import dbt_version
            if dbt_version < 'v1.8':
                from dbt.exceptions import DbtDatabaseError
            else:
                from dbt_common.exceptions import DbtDatabaseError
            if isinstance(e, DbtDatabaseError):
                if str(e).find('100051') >= 0:
                    # Snowflake error '100051 (22012): Division by zero"'
                    e = RecceException('No profile diff result due to the model is empty.', False)
            raise e

    def cancel(self):
        super().cancel()

        if self.connection:
            from recce.adapter.dbt_adapter import DbtAdapter
            dbt_adapter: DbtAdapter = default_context().adapter
            with dbt_adapter.connection_named("cancel"):
                dbt_adapter.cancel(self.connection)


class ProfileDiffResultDiffer(TaskResultDiffer):
    def _check_result_changed_fn(self, result):
        return self.diff(result['base'], result['current'])


class ProfileDiffCheckValidator(CheckValidator):

    def validate_check(self, check: Check):
        try:
            ProfileDiffParams(**check.params)
        except Exception as e:
            raise ValueError(f"Invalid check: {str(e)}")



import asyncio
import os
from pathlib import Path
from typing import List

import click
import uvicorn

from recce import event
from recce.artifact import upload_dbt_artifacts, download_dbt_artifacts
from recce.config import RecceConfig, RECCE_CONFIG_FILE, RECCE_ERROR_LOG_FILE
from recce.git import current_branch, current_default_branch
from recce.run import cli_run, check_github_ci_env
from recce.state import RecceStateLoader, RecceCloudStateManager
from recce.summary import generate_markdown_summary
from recce.util.logger import CustomFormatter
from recce.util.recce_cloud import RecceCloudException, get_recce_cloud_onboarding_state
from .core import RecceContext
from .event.track import TrackCommand

event.init()


def create_state_loader(review_mode, cloud_mode, state_file, cloud_options):
    from rich.console import Console
    console = Console()

    try:
        return RecceStateLoader(review_mode=review_mode, cloud_mode=cloud_mode,
                                state_file=state_file, cloud_options=cloud_options)
    except RecceCloudException as e:
        console.print("[[red]Error[/red]] Failed to load recce state file")
        console.print(f"Reason: {e.reason}")
        exit(1)
    except Exception as e:
        console.print("[[red]Error[/red]] Failed to load recce state file")
        console.print(f"Reason: {e}")
        exit(1)


def handle_debug_flag(**kwargs):
    if kwargs.get('debug'):
        import logging
        ch = logging.StreamHandler()
        ch.setFormatter(CustomFormatter())
        logging.basicConfig(handlers=[ch], level=logging.DEBUG)


def add_options(options):
    def _add_options(func):
        for option in reversed(options):
            func = option(func)
        return func

    return _add_options


dbt_related_options = [
    click.option('--target', '-t', help='Which target to load for the given profile.', type=click.STRING),
    click.option('--profile', help='Which existing profile to load.', type=click.STRING),
    click.option('--project-dir', help='Which directory to look in for the dbt_project.yml file.', type=click.Path(),
                 envvar="DBT_PROJECT_DIR"),
    click.option('--profiles-dir', help='Which directory to look in for the profiles.yml file.', type=click.Path(),
                 envvar="DBT_PROFILES_DIR"),
]

sqlmesh_related_options = [
    click.option('--sqlmesh', is_flag=True, help='Use SQLMesh ', hidden=True),
    click.option('--sqlmesh-envs', is_flag=False, help='SQLMesh envs to compare. SOURCE:TARGET', hidden=True),
    click.option('--sqlmesh-config', is_flag=False, help='SQLMesh config name to use', hidden=True),
]

recce_options = [
    click.option('--config', help='Path to the recce config file.', type=click.Path(), default=RECCE_CONFIG_FILE,
                 show_default=True),
    click.option('--error-log', help='Path to the error log file.', type=click.Path(), default=RECCE_ERROR_LOG_FILE,
                 hidden=True),
    click.option('--debug', is_flag=True, help='Enable debug mode.', hidden=True),
]

recce_cloud_options = [
    click.option('--cloud', is_flag=True, help='Fetch the state file from cloud.'),
    click.option('--cloud-token', help='The token used by Recce Cloud.', type=click.STRING,
                 envvar='GITHUB_TOKEN'),
    click.option('--state-file-host', help='The host to fetch the state file from.', type=click.STRING,
                 envvar='RECCE_STATE_FILE_HOST', default='', hidden=True),
    click.option('--password', '-p', help='The password to encrypt the state file in cloud.', type=click.STRING,
                 envvar='RECCE_STATE_PASSWORD'),
]

recce_dbt_artifact_dir_options = [
    click.option('--target-path', help='dbt artifacts directory for your development branch.',
                 type=click.STRING, default='target'),
    click.option('--target-base-path', help='dbt artifacts directory to be used as the base for the comparison.',
                 type=click.STRING, default='target-base'),
]


def _execute_sql(context, sql_template, base=False):
    try:
        import pandas as pd
    except ImportError:
        print("'pandas' package not found. You can install it using the command: 'pip install pandas'.")
        exit(1)

    from recce.adapter.dbt_adapter import DbtAdapter
    dbt_adapter: DbtAdapter = context.adapter
    with dbt_adapter.connection_named('recce'):
        sql = dbt_adapter.generate_sql(sql_template, base)
        response, result = dbt_adapter.execute(sql, fetch=True, auto_begin=True)
        table = result
        df = pd.DataFrame([row.values() for row in table.rows], columns=table.column_names)
        return df


@click.group()
@click.pass_context
def cli(ctx, **kwargs):
    """Recce: Data validation toolkit for comprehensive PR review"""


@cli.command(cls=TrackCommand)
def version():
    """
    Show version information
    """
    from recce import __version__
    print(__version__)


@cli.command(hidden=True, cls=TrackCommand)
@click.option('--sql', help='Sql template to query', required=True)
@click.option('--base', is_flag=True, help='Run the query on the base environment')
@add_options(dbt_related_options)
def query(sql, base: bool = False, **kwargs):
    """
    Run a query on the current or base environment

    Examples:\n

    - run an adhoc query\n
        recce query --sql 'select * from {{ ref("mymodel") }} order by 1'

    - run an adhoc query on base environment\n
        recce query --base --sql 'select * from {{ ref("mymodel") }} order by 1'
    """
    context = RecceContext.load(**kwargs)
    result = _execute_sql(context, sql, base=base)
    print(result.to_string(na_rep='-', index=False))


def _split_comma_separated(ctx, param, value):
    return value.split(',') if value else None


@cli.command(hidden=True, cls=TrackCommand)
@click.option('--sql', help='Sql template to query.', required=True)
@click.option('--primary-keys', type=click.STRING, help='Comma-separated list of primary key columns.',
              callback=_split_comma_separated)
@click.option('--keep-shape', is_flag=True, help='Keep unchanged columns. Otherwise, unchanged columns are hidden.')
@click.option('--keep-equal', is_flag=True,
              help='Keep values that are equal. Otherwise, equal values are shown as "-".')
@add_options(dbt_related_options)
def diff(sql, primary_keys: List[str] = None, keep_shape: bool = False, keep_equal: bool = False, **kwargs):
    """
    Run queries on base and current environments and diff the results

    Examples:\n

    - run adhoc queries and diff the results\n
        recce diff --sql 'select * from {{ ref("mymodel") }} order by 1'
    """

    context = RecceContext.load(**kwargs)
    before = _execute_sql(context, sql, base=True)
    if primary_keys is not None:
        before.set_index(primary_keys, inplace=True)
    after = _execute_sql(context, sql, base=False)
    if primary_keys is not None:
        after.set_index(primary_keys, inplace=True)

    before_aligned, after_aligned = before.align(after)
    diff = before_aligned.compare(after_aligned,
                                  result_names=('base', 'current'),
                                  keep_equal=keep_equal,
                                  keep_shape=keep_shape)
    print(diff.to_string(na_rep='-') if not diff.empty else 'no changes')


@cli.command(cls=TrackCommand)
@click.argument('state_file', required=False)
@click.option('--host', default='localhost', show_default=True, help='The host to bind to.')
@click.option('--port', default=8000, show_default=True, help='The port to bind to.', type=int)
@click.option('--review', is_flag=True, help='Open the state file in the review mode.')
@add_options(dbt_related_options)
@add_options(sqlmesh_related_options)
@add_options(recce_options)
@add_options(recce_dbt_artifact_dir_options)
@add_options(recce_cloud_options)
def server(host, port, state_file=None, **kwargs):
    """
    Launch the recce server

    STATE_FILE is the path to the recce state file. Defaults=None, which will be no persistent state.

    Examples:\n

    \b
    # Launch the recce server
    recce server

    \b
    # Launch the recce server with a state file
    recce server recce_state.json

    \b
    # Launch the server in the review mode
    recce server --review recce_state.json

    \b
    # Launch the server and synchronize the state with the cloud
    recce server --cloud
    recce server --review --cloud

    """

    from .server import app, AppState
    from rich.console import Console

    RecceConfig(config_file=kwargs.get('config'))

    handle_debug_flag(**kwargs)
    is_review = kwargs.get('review', False)
    is_cloud = kwargs.get('cloud', False)
    console = Console()
    cloud_options = None
    flag = {
        'show_onboarding_guide': True,
        'single_env_onboarding': False,
        'show_relaunch_hint': False
    }
    if is_cloud:
        cloud_options = {
            'host': kwargs.get('state_file_host'),
            'token': kwargs.get('cloud_token'),
            'password': kwargs.get('password'),
        }
        cloud_onboarding_state = get_recce_cloud_onboarding_state(kwargs.get('cloud_token'))
        flag['show_onboarding_guide'] = False if cloud_onboarding_state == 'completed' else True

    if not os.path.isdir(kwargs.get('target_base_path')):
        # Mark as single env onboarding mode if user provides the target-path only
        flag['single_env_onboarding'] = True
        flag['show_relaunch_hint'] = True
        target_path = kwargs.get('target_path')
        target_base_path = kwargs.get('target_base_path')
        # Use the target path as the base path
        kwargs['target_base_path'] = target_path

        # Show warning message
        console.rule('Warning', style='orange3')
        console.print(
            f'The folder "{target_base_path}" is not provided. '
            f'Will use the current target folder "{target_path}" as the base.\n')
        console.print('To setup your base environment, please refer to the following link for more information')
        console.print('https://datarecce.io/docs/get-started/#prepare-dbt-artifacts')
        console.print()

    state_loader = create_state_loader(is_review, is_cloud, state_file, cloud_options)

    if not state_loader.verify():
        error, hint = state_loader.error_and_hint
        console.print(f"[[red]Error[/red]] {error}")
        console.print(f"{hint}")
        exit(1)

    if state_loader.review_mode is True:
        console.rule("Recce Server : Review Mode")
    else:
        console.rule("Recce Server")

    result, message = RecceContext.verify_required_artifacts(**kwargs)
    if not result:
        console.print(f"[[red]Error[/red]] {message}")
        exit(1)

    state = AppState(state_loader=state_loader, kwargs=kwargs, flag=flag)
    app.state = state

    uvicorn.run(app, host=host, port=port, lifespan='on')


DEFAULT_RECCE_STATE_FILE = 'recce_state.json'


@cli.command(cls=TrackCommand)
@click.option('-o', '--output', help='Path of the output state file.', type=click.Path(),
              default=DEFAULT_RECCE_STATE_FILE, show_default=True)
@click.option('--state-file', help='Path of the import state file.', type=click.Path())
@click.option('--summary', help='Path of the summary markdown file.', type=click.Path())
@click.option('--skip-query', is_flag=True, help='Skip running the queries for the checks.')
@click.option('--git-current-branch', help='The git branch of the current environment.', type=click.STRING,
              envvar='GITHUB_HEAD_REF')
@click.option('--git-base-branch', help='The git branch of the base environment.', type=click.STRING,
              envvar='GITHUB_BASE_REF')
@click.option('--github-pull-request-url', help='The github pull request url to use for the lineage.',
              type=click.STRING)
@add_options(dbt_related_options)
@add_options(sqlmesh_related_options)
@add_options(recce_options)
@add_options(recce_dbt_artifact_dir_options)
@add_options(recce_cloud_options)
def run(output, **kwargs):
    """
    Run recce and output the state file

    Examples:\n

    \b
    # Run recce and output to the default path [recce_state.json]
    recce run

    \b
    # Run recce and output to the specified path
    recce run -o my_recce_state.json

    \b
    # Run recce and output to the specified path
    recce run --cloud --cloud-token <token> --password <password>

    """
    from rich.console import Console
    handle_debug_flag(**kwargs)
    console = Console()
    is_github_action, pr_url = check_github_ci_env(**kwargs)
    if is_github_action is True and pr_url is not None:
        kwargs['github_pull_request_url'] = pr_url

    # Initialize Recce Config
    RecceConfig(config_file=kwargs.get('config'))

    cloud_mode = kwargs.get('cloud', False)
    state_file = kwargs.get('state_file')
    cloud_options = {
        'host': kwargs.get('state_file_host'),
        'token': kwargs.get('cloud_token'),
        'password': kwargs.get('password'),
    } if cloud_mode else None

    state_loader = create_state_loader(review_mode=False, cloud_mode=cloud_mode, state_file=state_file,
                                       cloud_options=cloud_options)

    if not state_loader.verify():
        error, hint = state_loader.error_and_hint
        console.print(f"[[red]Error[/red]] {error}")
        console.print(f"{hint}")
        exit(1)

    result, message = RecceContext.verify_required_artifacts(**kwargs)
    if not result:
        console.print(f"[[red]Error[/red]] {message}")
        exit(1)

    # Verify the output state file path
    try:
        if os.path.isdir(output) or output.endswith('/'):

            output_dir = Path(output)
            # Create the directory if not exists
            output_dir.mkdir(parents=True, exist_ok=True)
            output = os.path.join(output, DEFAULT_RECCE_STATE_FILE)
            console.print(
                f"[[yellow]Warning[/yellow]] The path '{output_dir}' is a directory. "
                f"The state file will be saved as '{output}'.")
        else:
            # Create the parent directory if not exists
            output_dir = Path(output).parent
            output_dir.mkdir(parents=True, exist_ok=True)
    except FileExistsError as e:
        console.print(f"[[red]Error[/red]] Failed to access file path '{output}'.")
        console.print(f"Reason: {e}")
        exit(1)

    return asyncio.run(cli_run(output, state_loader=state_loader, **kwargs))


@cli.command(cls=TrackCommand)
@click.argument('state_file', required=False)
@click.option('--format', '-f', help='Output format. Currently only markdown is supported.',
              type=click.Choice(['markdown', 'mermaid', 'check'], case_sensitive=False),
              default='markdown', show_default=True, hidden=True)
@add_options(dbt_related_options)
@add_options(recce_options)
@add_options(recce_cloud_options)
def summary(state_file, **kwargs):
    """
        Generate a summary of the recce state file
    """
    from rich.console import Console
    from .core import load_context
    handle_debug_flag(**kwargs)
    console = Console()
    cloud_mode = kwargs.get('cloud', False)
    cloud_options = {
        'host': kwargs.get('state_file_host'),
        'token': kwargs.get('cloud_token'),
        'password': kwargs.get('password'),
    } if cloud_mode else None

    state_loader = create_state_loader(review_mode=True, cloud_mode=cloud_mode, state_file=state_file,
                                       cloud_options=cloud_options)

    if not state_loader.verify():
        error, hint = state_loader.error_and_hint
        console.print(f"[[red]Error[/red]] {error}")
        console.print(f"{hint}")
        exit(1)
    try:
        # Load context in review mode, won't need to check dbt_project.yml file.
        ctx = load_context(**kwargs, state_loader=state_loader, review=True)
    except Exception as e:
        console.print("[[red]Error[/red]] Failed to generate summary")
        console.print(f"{e}")
        exit(1)

    output = generate_markdown_summary(ctx, summary_format=kwargs.get('format'))
    print(output)


@cli.group('cloud', short_help='Manage Recce Cloud state file.')
def cloud(**kwargs):
    # Manage Recce Cloud.
    pass


@cloud.command(cls=TrackCommand)
@click.option('--cloud-token', help='The token used by Recce Cloud.', type=click.STRING,
              envvar='GITHUB_TOKEN')
@click.option('--state-file-host', help='The host to fetch the state file from.', type=click.STRING,
              envvar='RECCE_STATE_FILE_HOST', default='', hidden=True)
@click.option('--password', '-p', help='The password to encrypt the state file in cloud.', type=click.STRING,
              envvar='RECCE_STATE_PASSWORD')
@click.option('--force', '-f', help='Bypasses the confirmation prompt. Purge the state file directly.', is_flag=True)
@add_options(recce_options)
def purge(**kwargs):
    """
        Purge the state file from cloud
    """
    from rich.console import Console
    handle_debug_flag(**kwargs)
    console = Console()
    state_loader = None
    cloud_options = {
        'host': kwargs.get('state_file_host'),
        'token': kwargs.get('cloud_token'),
        'password': kwargs.get('password'),
    }
    force_to_purge = kwargs.get('force', False)

    try:
        console.rule('Check Recce State from Cloud')
        state_loader = RecceStateLoader(review_mode=False, cloud_mode=True,
                                        state_file=None, cloud_options=cloud_options)
    except Exception:
        console.print("[[yellow]Skip[/yellow]] Cannot access existing state file from cloud. Purge it directly.")

    if state_loader is None:
        try:
            if force_to_purge is True or click.confirm('\nDo you want to purge the state file?'):
                rc, err_msg = RecceCloudStateManager(cloud_options).purge_cloud_state()
                if rc is True:
                    console.rule('Purged Successfully')
                else:
                    console.rule('Failed to Purge', style='red')
                    console.print(f'Reason: {err_msg}')

        except click.exceptions.Abort:
            pass
        return 0

    info = state_loader.info()
    if info is None:
        console.print("[[yellow]Skip[/yellow]] No state file found in cloud.")
        return 0

    pr_info = info.get('pull_request')
    console.print('[green]State File hosted by[/green]', info.get('source'))
    console.print('[green]GitHub Repository[/green]', info.get('pull_request').repository)
    console.print(f'[green]GitHub Pull Request[/green]\n{pr_info.title} #{pr_info.id}')
    console.print(f'Branch merged into [blue]{pr_info.base_branch}[/blue] from [blue]{pr_info.branch}[/blue]')
    console.print(pr_info.url)

    try:
        if force_to_purge is True or click.confirm('\nDo you want to purge the state file?'):
            response = state_loader.purge()
            if response is True:
                console.rule('Purged Successfully')
            else:
                console.rule('Failed to Purge', style='red')
                console.print(f'Reason: {state_loader.error_message}')
    except click.exceptions.Abort:
        pass

    return 0


@cloud.command(cls=TrackCommand)
@click.argument('state_file', type=click.Path(exists=True))
@click.option('--cloud-token', help='The token used by Recce Cloud.', type=click.STRING,
              envvar='GITHUB_TOKEN')
@click.option('--state-file-host', help='The host to fetch the state file from.', type=click.STRING,
              envvar='RECCE_STATE_FILE_HOST', default='', hidden=True)
@click.option('--password', '-p', help='The password to encrypt the state file in cloud.', type=click.STRING,
              envvar='RECCE_STATE_PASSWORD')
@add_options(recce_options)
def upload(state_file, **kwargs):
    """
        Upload the state file to cloud
    """
    from rich.console import Console

    handle_debug_flag(**kwargs)
    cloud_options = {
        'host': kwargs.get('state_file_host'),
        'token': kwargs.get('cloud_token'),
        'password': kwargs.get('password'),
    }

    console = Console()

    # load local state
    state_loader = create_state_loader(review_mode=False, cloud_mode=False, state_file=state_file,
                                       cloud_options=cloud_options)

    if not state_loader.verify():
        error, hint = state_loader.error_and_hint
        console.print(f"[[red]Error[/red]] {error}")
        console.print(f"{hint}")
        exit(1)

    # check if state exists in cloud
    state_manager = RecceCloudStateManager(cloud_options)
    cloud_state_file_exists = state_manager.check_cloud_state_exists()

    if cloud_state_file_exists and not click.confirm('\nDo you want to overwrite the existing state file?'):
        return 0

    console.print(state_manager.upload_state_to_cloud(state_loader.state))


@cloud.command(cls=TrackCommand)
@click.option('-o', '--output', help='Path of the downloaded state file.', type=click.STRING,
              default=DEFAULT_RECCE_STATE_FILE, show_default=True)
@click.option('--cloud-token', help='The token used by Recce Cloud.', type=click.STRING,
              envvar='GITHUB_TOKEN')
@click.option('--state-file-host', help='The host to fetch the state file from.', type=click.STRING,
              envvar='RECCE_STATE_FILE_HOST', default='', hidden=True)
@click.option('--password', '-p', help='The password to encrypt the state file in cloud.', type=click.STRING,
              envvar='RECCE_STATE_PASSWORD')
@add_options(recce_options)
def download(**kwargs):
    """
        Download the state file to cloud
    """
    from rich.console import Console

    handle_debug_flag(**kwargs)
    filepath = kwargs.get('output')
    cloud_options = {
        'host': kwargs.get('state_file_host'),
        'token': kwargs.get('cloud_token'),
        'password': kwargs.get('password'),
    }

    console = Console()

    # check if state exists in cloud
    state_manager = RecceCloudStateManager(cloud_options)
    cloud_state_file_exists = state_manager.check_cloud_state_exists()

    if not cloud_state_file_exists:
        console.print('[yellow]Skip[/yellow] No state file found in cloud.')
        return 0

    state_manager.download_state_from_cloud(filepath)
    console.print(f'Downloaded state file to "{filepath}"')


@cloud.command(cls=TrackCommand)
@click.option('--cloud-token', help='The token used by Recce Cloud.', type=click.STRING,
              envvar='GITHUB_TOKEN')
@click.option('--branch', '-b', help='The branch of the provided artifacts.', type=click.STRING,
              envvar='GITHUB_HEAD_REF', default=current_branch(), show_default=True)
@click.option('--target-path', help='dbt artifacts directory for your artifacts.', type=click.STRING, default='target',
              show_default=True)
@click.option('--password', '-p', help='The password to encrypt the dbt artifacts in cloud.', type=click.STRING,
              envvar='RECCE_STATE_PASSWORD', required=True)
@add_options(recce_options)
def upload_artifacts(**kwargs):
    """
        Upload the dbt artifacts to cloud

        Upload the dbt artifacts (metadata.json, catalog.json) to Recce Cloud for the given branch.
        The password is used to encrypt the dbt artifacts in the cloud. You will need the password to download the dbt artifacts.

        By default, the artifacts are uploaded to the current branch. You can specify the branch using the --branch option.
        The target path is set to 'target' by default. You can specify the target path using the --target-path option.
    """
    from rich.console import Console
    console = Console()
    cloud_token = kwargs.get('cloud_token')
    password = kwargs.get('password')
    target_path = kwargs.get('target_path')
    branch = kwargs.get('branch')

    try:
        rc = upload_dbt_artifacts(target_path, branch=branch,
                                  token=cloud_token, password=password,
                                  debug=kwargs.get('debug', False))
        console.rule('Uploaded Successfully')
        console.print(
            f'Uploaded dbt artifacts to Recce Cloud for branch "{branch}" from "{os.path.abspath(target_path)}"')
    except Exception as e:
        console.rule('Failed to Upload', style='red')
        console.print("[[red]Error[/red]] Failed to upload the dbt artifacts to cloud.")
        console.print(f"Reason: {e}")
        rc = 1
    return rc


def _download_artifacts(branch, cloud_token, console, kwargs, password, target_path):
    try:
        rc = download_dbt_artifacts(target_path, branch=branch, token=cloud_token, password=password,
                                    force=kwargs.get('force', False),
                                    debug=kwargs.get('debug', False))
        console.rule('Downloaded Successfully')
        console.print(
            f'Downloaded dbt artifacts from Recce Cloud for branch "{branch}" to "{os.path.abspath(target_path)}"')
    except Exception as e:
        console.rule('Failed to Download', style='red')
        console.print("[[red]Error[/red]] Failed to download the dbt artifacts from cloud.")
        reason = str(e)

        if 'Requests specifying Server Side Encryption with Customer provided keys must provide the correct secret key' in reason:
            console.print("Reason: Decryption failed due to incorrect password.")
            console.print(
                "Please provide the correct password to decrypt the dbt artifacts. Or re-upload the dbt artifacts with a new password.")
        elif 'The specified key does not exist' in reason:
            console.print("Reason: The dbt artifacts is not found in the cloud.")
            console.print("Please upload the dbt artifacts to the cloud before downloading it.")
        else:
            console.print(f"Reason: {reason}")
        rc = 1
    return rc


@cloud.command(cls=TrackCommand)
@click.option('--cloud-token', help='The token used by Recce Cloud.', type=click.STRING,
              envvar='GITHUB_TOKEN')
@click.option('--branch', '-b', help='The branch of the selected artifacts.', type=click.STRING,
              envvar='GITHUB_BASE_REF', default=current_branch(), show_default=True)
@click.option('--target-path', help='The dbt artifacts directory for your artifacts.', type=click.STRING,
              default='target', show_default=True)
@click.option('--password', '-p', help='The password to decrypt the dbt artifacts in cloud.', type=click.STRING,
              envvar='RECCE_STATE_PASSWORD', required=True)
@click.option('--force', '-f', help='Bypasses the confirmation prompt. Download the artifacts directly.',
              is_flag=True)
@add_options(recce_options)
def download_artifacts(**kwargs):
    """
        Download the dbt artifacts from cloud

        Download the dbt artifacts (metadata.json, catalog.json) from Recce Cloud for the given branch.
        The password is used to decrypt the dbt artifacts in the cloud.

        By default, the artifacts are downloaded from the current branch. You can specify the branch using the --branch option.
        The target path is set to 'target' by default. You can specify the target path using the --target-path option.
    """
    from rich.console import Console
    console = Console()
    cloud_token = kwargs.get('cloud_token')
    password = kwargs.get('password')
    target_path = kwargs.get('target_path')
    branch = kwargs.get('branch')
    return _download_artifacts(branch, cloud_token, console, kwargs, password, target_path)


@cloud.command(cls=TrackCommand)
@click.option('--cloud-token', help='The token used by Recce Cloud.', type=click.STRING,
              envvar='GITHUB_TOKEN')
@click.option('--branch', '-b', help='The branch of the selected artifacts.', type=click.STRING,
              envvar='GITHUB_BASE_REF', default=current_default_branch(), show_default=True)
@click.option('--target-path', help='The dbt artifacts directory for your artifacts.', type=click.STRING,
              default='target-base', show_default=True)
@click.option('--password', '-p', help='The password to decrypt the dbt artifacts in cloud.', type=click.STRING,
              envvar='RECCE_STATE_PASSWORD', required=True)
@click.option('--force', '-f', help='Bypasses the confirmation prompt. Download the artifacts directly.',
              is_flag=True)
@add_options(recce_options)
def download_base_artifacts(**kwargs):
    """
        Download the base dbt artifacts from cloud

        Download the base dbt artifacts (metadata.json, catalog.json) from Recce Cloud.
        This is useful when you start to set up the base dbt artifacts for the first time.

        Please make sure you have uploaded the dbt artifacts before downloading them.
    """
    from rich.console import Console
    console = Console()
    cloud_token = kwargs.get('cloud_token')
    password = kwargs.get('password')
    target_path = kwargs.get('target_path')
    branch = kwargs.get('branch')
    return _download_artifacts(branch, cloud_token, console, kwargs, password, target_path)


@cli.group('github', short_help='GitHub related commands', hidden=True)
def github(**kwargs):
    pass


@github.command(cls=TrackCommand,
                short_help='Download the artifacts from the GitHub repository based on the current Pull Request.')
@click.option('--github-token', help='The github token to use for accessing GitHub repo.', type=click.STRING,
              envvar='GITHUB_TOKEN')
@click.option('--github-repo', help='The github repo to use for accessing GitHub repo.', type=click.STRING,
              envvar='GITHUB_REPOSITORY')
def artifact(**kwargs):
    from recce.github import recce_ci_artifact
    return recce_ci_artifact(**kwargs)


if __name__ == "__main__":
    cli()



import os
import sys
from typing import List, Dict, Set, Union, Type, Optional
from uuid import UUID

from pydantic import BaseModel

from recce.apis.check_func import get_node_name_by_id
from recce.core import RecceContext
from recce.models import CheckDAO, RunDAO, RunType, Run
from recce.tasks.core import TaskResultDiffer
from recce.tasks.histogram import HistogramDiffTaskResultDiffer
from recce.tasks.profile import ProfileDiffResultDiffer
from recce.tasks.query import QueryDiffResultDiffer
from recce.tasks.rowcount import RowCountDiffResultDiffer
from recce.tasks.schema import SchemaDiffResultDiffer
from recce.tasks.top_k import TopKDiffTaskResultDiffer
from recce.tasks.valuediff import ValueDiffTaskResultDiffer, ValueDiffDetailTaskResultDiffer

RECCE_CLOUD_HOST = os.environ.get('RECCE_CLOUD_HOST', 'https://cloud.datarecce.io')

ADD_COLOR = '#1dce00'
MODIFIED_COLOR = '#ffa502'
REMOVE_COLOR = '#ff067e'

MAX_MERMAID_TEXT_SIZE = 50000  # source: https://mermaid.js.org/config/schema-docs/config.html#maxtextsize


def _warn(msg):
    # print to stderr
    print(f"warning: {msg}", file=sys.stderr)


class Node:
    id: str
    name: str
    data_from: str
    resource_type: str
    package_name: str
    parents: List[str]
    children: List[str]
    base_data: dict
    current_data: dict

    def __init__(self, node_id: str, node_data: dict, data_from: str = 'base'):
        self.id = node_id
        self.name = node_data['name']
        self.data_from = data_from
        self.resource_type = node_data['resource_type']
        self.package_name = node_data['package_name']
        self.children = []
        self.parents = []

        self.base_data = {}
        self.current_data = {}

        if data_from == 'base':
            self.base_data = node_data
        elif data_from == 'current':
            self.current_data = node_data

    @property
    def change_status(self):
        base_checksum = self.base_data.get('checksum', {}).get('checksum')
        curr_checksum = self.current_data.get('checksum', {}).get('checksum')
        if self.data_from == 'base':
            return 'removed'
        elif self.data_from == 'current':
            return 'added'
        elif base_checksum and curr_checksum and base_checksum != curr_checksum:
            return 'modified'
        return None

    def update_data(self, node_data: dict, data_from: str):
        if data_from not in ['base', 'current']:
            raise ValueError(f'Invalid data_from value: {data_from}')
        if self.data_from != data_from:
            self.data_from = 'both'

            if data_from == 'base':
                self.base_data = node_data
            elif data_from == 'current':
                self.current_data = node_data

    def append_parent(self, parent_id: str):
        if parent_id not in self.parents:
            self.parents.append(parent_id)

    def append_child(self, child_id: str):
        if child_id not in self.children:
            self.children.append(child_id)

    def _cal_row_count_delta_percentage(self):
        row_count_diff, run_result = _get_node_row_count_diff(self.id, self.name)
        if row_count_diff:
            base = run_result.get('base', 0)
            current = run_result.get('curr', 0)
            if int(current) > int(base):
                p = (int(current) - int(base)) / int(current) * 100
                return f'🔼 +{round(p, 2) if p > 0.1 else "<0.1"}%'
            else:
                p = (int(base) - int(current)) / int(current) * 100
                return f'🔽 -{round(p, 2) if p > 0.1 else "<0.1"}%'
        return None

    def _get_schema_diff(self):
        base_schema = self.base_data.get('columns', {})
        current_schema = self.current_data.get('columns', {})
        schema_diff = TaskResultDiffer.diff(base_schema, current_schema)
        return schema_diff

    def _what_changed(self, checks=None):
        changes = []
        if self.change_status == 'added':
            return ['Added Node']
        elif self.change_status == 'removed':
            return ['Removed Node']
        elif self.change_status == 'modified':
            changes.append('Code')
        row_count_delta_percentage = self._cal_row_count_delta_percentage()
        if row_count_delta_percentage:
            changes.append(f'Row Count {row_count_delta_percentage}')
        schema_diff = self._get_schema_diff()
        if schema_diff:
            changes.append('Schema')

        if checks:
            for check in checks:
                check_type = check.type
                if check_type == RunType.ROW_COUNT_DIFF or check_type == RunType.SCHEMA_DIFF:
                    # Skip the row count and schema diff check, since we already have it.
                    continue
                if check.node_ids and self.id in check.node_ids:
                    changes.append(str(check.type).replace('_', ' ').title())
        return changes

    def get_node_str(self, checks=None):
        is_changed = False
        style = None

        if self.change_status is not None:
            is_changed = True
            if self.change_status == 'added':
                style = f'style {self.id} stroke:{ADD_COLOR}'
            elif self.change_status == 'modified':
                style = f'style {self.id} stroke:{MODIFIED_COLOR}'
            elif self.change_status == 'removed':
                style = f'style {self.id} stroke:{REMOVE_COLOR}'

        if checks:
            for check in checks:
                if check.node_ids and self.id in check.node_ids:
                    is_changed = True

        content_output = f'{self.id}["{self.name}'
        if is_changed:
            content_output += '\n\n[What\'s Changed]\n'
            changes = self._what_changed(checks)
            content_output += ', '.join(changes)

        content_output += '"]\n'
        if style:
            content_output += f'{style}\n'
        return content_output


class Edge:
    id: str
    edge_from: str
    child_id: str
    parent_id: str
    change_status: Union[str, None]

    def __init__(self, edge_id: str, parent_id: str, child_id: str, edge_from: str = 'base'):
        self.id = edge_id
        self.edge_from = edge_from
        self.child_id = child_id
        self.parent_id = parent_id

    def update_edge_from(self, edge_from: str):
        if self.edge_from != edge_from:
            self.edge_from = 'both'


class CheckSummary(BaseModel):
    id: UUID
    type: RunType
    name: str
    description: str
    changes: dict
    node_ids: Optional[List[str]]
    changed_nodes: Optional[List[str]]

    @property
    def related_nodes(self):
        nodes = []

        for node_id in self.node_ids or []:
            name = get_node_name_by_id(node_id)
            nodes.append(name)

        return nodes


check_result_differ_registry: Dict[RunType, Type[TaskResultDiffer]] = {
    RunType.VALUE_DIFF: ValueDiffTaskResultDiffer,
    RunType.VALUE_DIFF_DETAIL: ValueDiffDetailTaskResultDiffer,
    RunType.ROW_COUNT_DIFF: RowCountDiffResultDiffer,
    RunType.QUERY_DIFF: QueryDiffResultDiffer,
    RunType.TOP_K_DIFF: TopKDiffTaskResultDiffer,
    RunType.HISTOGRAM_DIFF: HistogramDiffTaskResultDiffer,
    RunType.PROFILE_DIFF: ProfileDiffResultDiffer,
}


def differ_factory(run: Run):
    differ_clz = check_result_differ_registry.get(run.type)
    if not differ_clz:
        raise NotImplementedError()
    return differ_clz(run)


class LineageGraph:
    nodes: Dict[str, Node] = {}
    edges: Dict[str, Edge] = {}
    checks: List[CheckSummary] = None

    def create_node(self, node_id: str, node_data: dict, data_from: str = 'base'):
        if node_id not in self.nodes:
            self.nodes[node_id] = Node(node_id, node_data, data_from)
        else:
            self.nodes[node_id].update_data(node_data, data_from)

    def create_edge(self, parent_id: str, child_id: str, edge_from: str = 'base'):
        if parent_id not in self.nodes:
            _warn(f'Parent node {parent_id} not found in graph')
            return
        if child_id not in self.nodes:
            _warn(f'Child node {child_id} not found in graph')
            return

        edge_id = f'{parent_id}-->{child_id}'
        if edge_id in self.edges:
            self.edges[edge_id].update_edge_from(edge_from)
        else:
            self.edges[edge_id] = Edge(edge_id, parent_id, child_id, edge_from)
            self.nodes[parent_id].append_child(child_id)
            self.nodes[child_id].append_parent(parent_id)

    @property
    def modified_set(self) -> Set[str]:
        return set([node_id for node_id, node in self.nodes.items() if node.change_status == 'modified'])

    def get_edge_str(self, edge_id):
        edge = self.edges[edge_id]
        child = self.nodes[edge.child_id]

        if child.change_status == 'removed':
            return f'{edge.parent_id}-.->{edge.child_id}\n'
        if child.change_status is None or child.change_status == 'modified':
            return f'{edge.parent_id}---->{edge.child_id}\n'
        if child.change_status == 'added':
            return f'{edge.parent_id}-...->{edge.child_id}\n'


def _build_lineage_graph(base, current) -> LineageGraph:
    graph = LineageGraph()

    # Init Graph nodes with base & current nodes
    for node_id, node_data in base.get('nodes', {}).items():
        graph.create_node(node_id, node_data, 'base')

    for node_id, node_data in current.get('nodes', {}).items():
        if node_id not in graph.nodes:
            node = Node(node_id, node_data, 'current')
            graph.nodes[node_id] = node
        else:
            node = graph.nodes[node_id]
            node.update_data(node_data, 'current')

    # Build edges
    for child_id, parents in base.get('parent_map', {}).items():
        for parent_id in parents:
            graph.create_edge(parent_id, child_id, 'base')
    for child_id, parents in current.get('parent_map', {}).items():
        for parent_id in parents:
            graph.create_edge(parent_id, child_id, 'current')

    return graph


def _build_node_schema(lineage, node_id):
    return lineage.get('nodes', {}).get(node_id, {}).get('columns', {})


def _get_node_row_count_diff(node_id, node_name):
    row_count_runs = RunDAO().list(type_filter=RunType.ROW_COUNT_DIFF)
    for run in row_count_runs:
        if node_id in run.params.get('node_ids', []):
            result = run.result.get(node_name, {})
            diff = TaskResultDiffer.diff(result.get('base'), result.get('curr'))
            return diff, result
        elif run.params.get('node_id') == node_id:
            result = run.result.get(node_name, {})
            diff = TaskResultDiffer.diff(result.get('base'), result.get('curr'))
            return diff, result
    return None, None


def _generate_mismatched_nodes_summary(check: CheckSummary, limit: int = 3) -> str:
    if not check.related_nodes:
        return 'N/A'

    nodes = check.related_nodes
    if check.changed_nodes:
        # the mismatch nodes within the related nodes (when apply node selection)
        # currently only schema_diff and row_count_diff are supported to apply node selection
        nodes = check.changed_nodes

    if len(nodes) <= limit:
        return ', '.join(nodes)

    display_nodes = nodes[:limit - 1]
    return ', '.join(display_nodes) + f', and {len(nodes) - len(display_nodes)} more nodes'


def generate_summary_metadata(base_lineage, curr_lineage):
    from py_markdown_table.markdown_table import markdown_table

    base_manifest = base_lineage.get('manifest_metadata')
    base_catalog = base_lineage.get('catalog_metadata')
    curr_manifest = curr_lineage.get('manifest_metadata')
    curr_catalog = curr_lineage.get('catalog_metadata')

    metadata = [
        {
            '': 'Base',
            'Manifest': base_manifest.generated_at.strftime('%Y-%m-%d %H:%M:%S'),
            'Catalog': base_catalog.generated_at.strftime('%Y-%m-%d %H:%M:%S') if base_catalog else 'N/A'
        },
        {
            '': 'Current',
            'Manifest': curr_manifest.generated_at.strftime('%Y-%m-%d %H:%M:%S'),
            'Catalog': curr_catalog.generated_at.strftime('%Y-%m-%d %H:%M:%S') if curr_catalog else 'N/A'
        }
    ]

    return markdown_table(metadata).set_params(
        quote=False,
        row_sep='markdown',
        padding_width=1,
        padding_weight='right'  # Aligns the cell's contents to the beginning of the cell
    ).get_markdown()


def generate_check_summary(base_lineage, curr_lineage) -> (List[CheckSummary], Dict[str, int]):
    runs = RunDAO().list()
    checks = CheckDAO().list()
    checks_summary: List[CheckSummary] = []
    failed_checks_count = 0

    # TODO: find a way to count failed checks, currently the state file won't include failed checks

    def _find_run_by_check_id(check_id):
        runs_for_check = [r for r in runs if str(check_id) == str(r.check_id)]
        if runs_for_check:
            return runs_for_check[-1]
        return None

    for check in checks:
        run = _find_run_by_check_id(check.check_id)
        differ = None
        if run is not None and run.error is not None:
            failed_checks_count += 1
            continue
        elif check.type == RunType.SCHEMA_DIFF:
            differ = SchemaDiffResultDiffer(check, base_lineage, curr_lineage)
        elif (check.type in [RunType.ROW_COUNT_DIFF, RunType.QUERY_DIFF,
                             RunType.VALUE_DIFF, RunType.VALUE_DIFF_DETAIL, RunType.PROFILE_DIFF,
                             RunType.TOP_K_DIFF, RunType.HISTOGRAM_DIFF] and run is not None):
            # Check the result is changed or not
            differ = differ_factory(run)

        if differ and differ.changes is not None:
            checks_summary.append(
                CheckSummary(
                    id=check.check_id,
                    type=check.type,
                    name=check.name,
                    description=check.description,
                    changes=differ.changes,
                    node_ids=differ.related_node_ids,
                    changed_nodes=differ.changed_nodes
                )
            )

    return checks_summary, {
        'total': len(checks),
        'mismatch': len(checks_summary),
        'failed': failed_checks_count,
    }


def generate_mermaid_lineage_graph(graph: LineageGraph):
    content = up_to_level_content = 'graph LR\n'
    is_not_modified = False
    # Only show the modified nodes and there children
    queue = list(graph.modified_set)
    display_nodes = set()
    display_edge = set()

    if len(queue) == 0:
        is_not_modified = True

    while len(queue) > 0:

        level_nodes = len(queue)
        for _ in range(level_nodes):
            node_id = queue.pop(0)
            if node_id in display_nodes:
                # Skip if already displayed
                continue

            display_nodes.add(node_id)
            node = graph.nodes[node_id]
            content += node.get_node_str(graph.checks)
            for child_id in node.children:
                queue.append(child_id)
                edge_id = f'{node_id}-->{child_id}'
                if edge_id not in display_edge:
                    display_edge.add(edge_id)
                    content += graph.get_edge_str(edge_id)

        if len(content) > MAX_MERMAID_TEXT_SIZE:
            break

        up_to_level_content = content

    return up_to_level_content, is_not_modified, len(content) > MAX_MERMAID_TEXT_SIZE


def generate_markdown_summary(ctx: RecceContext, summary_format: str = 'markdown'):
    curr_lineage = ctx.get_lineage(base=False)
    base_lineage = ctx.get_lineage(base=True)
    summary_metadata = generate_summary_metadata(base_lineage, curr_lineage)
    graph = _build_lineage_graph(base_lineage, curr_lineage)
    graph.checks, check_statistics = generate_check_summary(base_lineage, curr_lineage)
    mermaid_content, is_empty_graph, is_partial_graph = generate_mermaid_lineage_graph(graph)
    check_content = generate_check_content(graph, check_statistics)

    if summary_format == 'mermaid':
        return mermaid_content
    elif summary_format == 'check':
        return check_content
    elif summary_format == 'markdown':

        content = '# Recce Summary\n'
        content += f'## Manifest Information\n{summary_metadata}\n'

        if is_empty_graph is False:
            content += f'''
## Lineage Graph
{"_Too many nodes to generate! Please see the full lineage graph on Recce instance._" if is_partial_graph else ''}
```mermaid
{mermaid_content}
```
'''
        else:
            content += '''
## Lineage Graph
No changed module was detected.
'''
        if check_content:
            content += check_content

        if ctx.state_loader.cloud_mode:
            pr_info = ctx.state_loader.pr_info
            content += f'\nSee PR page: {RECCE_CLOUD_HOST}/{pr_info.repository}/pulls/{pr_info.id}\n'

        return content


def generate_check_content(graph, check_statistics):
    from py_markdown_table.markdown_table import markdown_table
    content = ''
    check_content = None
    # Generate the check summary if we found any changes
    if len(graph.checks) > 0:
        data = []
        for check in graph.checks:
            data.append({
                'Name': check.name,
                'Type': str(check.type).replace('_', ' ').title(),
                'Mismatched Nodes': _generate_mismatched_nodes_summary(check),
                # Temporarily remove the type of changes, until we implement a better way to display it.
                # 'Type of Changes': _formate_changes(check.changes)
            })
        check_content = markdown_table(data).set_params(
            quote=False,
            row_sep='markdown',
            padding_width=1,
            padding_weight='right'  # Aligns the cell's contents to the beginning of the cell
        ).get_markdown()

    if check_statistics.get('total', 0) > 0:
        warning_message = ''
        statistics = {
            'Checks Run': check_statistics.get('total', 0),
            'Data Mismatch Detected': check_statistics.get('mismatch', 0),
        }
        if check_statistics.get('failed', 0) > 0:
            statistics['Incomplete Checks'] = check_statistics.get('failed', 0)
            warning_message = '''
:warning: **Incomplete Checks** refers to checks that did not successfully run due to configuration or SQL errors.
Please check the output of `recce run` for more information
'''
        check_summary = markdown_table([statistics]).set_params(quote=False, row_sep='markdown').get_markdown()
        content += f'''
## Checks Summary
{check_summary}
{warning_message}
'''
    if check_content:
        content += f'''
### Checks of Data Mismatch Detected
{check_content}
'''
    return content



import pytest

from recce.tasks import QueryDiffTask, QueryTask


def test_query_diff_in_client(dbt_test_helper):
    csv_data_curr = """
        customer_id,name,age
        1,Alice,30
        2,Bob,25
        3,Charlie,35
        """

    csv_data_base = """
        customer_id,name,age
        1,Alice,35
        2,Bob,25
        3,Charlie,35
        """

    dbt_test_helper.create_model("customers", csv_data_base, csv_data_curr)
    params = dict(sql_template=f'select * from {{{{ ref("customers") }}}}')
    task = QueryTask(params)
    run_result = task.execute()
    assert len(run_result.current.data) == 3


def test_query_diff_in_client(dbt_test_helper):
    csv_data_curr = """
        customer_id,name,age
        1,Alice,30
        2,Bob,25
        3,Charlie,35
        """

    csv_data_base = """
        customer_id,name,age
        1,Alice,35
        2,Bob,25
        3,Charlie,35
        """

    dbt_test_helper.create_model("customers", csv_data_base, csv_data_curr)
    params = dict(sql_template=f'select * from {{{{ ref("customers") }}}}')
    task = QueryDiffTask(params)
    run_result = task.execute()
    assert len(run_result.base.data) == 3
    assert len(run_result.current.data) == 3

    params = dict(
        base_sql_template=f'select * from {{{{ ref("customers") }}}} where customer_id <= 2',
        sql_template=f'select * from {{{{ ref("customers") }}}}',
    )
    task = QueryDiffTask(params)
    run_result = task.execute()
    assert len(run_result.base.data) == 2
    assert len(run_result.current.data) == 3


def test_query_diff_in_warehouse(dbt_test_helper):
    csv_data_curr = """
        customer_id,name,age
        1,Alice,30
        2,Bob,25
        3,Charlie,35
        """

    csv_data_base = """
        customer_id,name,age
        1,Alice,35
        2,Bob,25
        3,Charlie,35
        """

    dbt_test_helper.create_model("customers", csv_data_base, csv_data_curr)
    params = dict(sql_template=f'select * from {{{{ ref("customers") }}}}', primary_keys=['customer_id'])
    task = QueryDiffTask(params)
    run_result = task.execute()
    assert len(run_result.diff.data) == 2

    params = dict(
        base_sql_template=f'select * from {{{{ ref("customers") }}}} where customer_id == 1',
        sql_template=f'select * from {{{{ ref("customers") }}}}',
        primary_keys=['customer_id'],
    )
    task = QueryDiffTask(params)
    run_result = task.execute()
    assert len(run_result.diff.data) == 4


def test_validator():
    from recce.tasks.query import QueryCheckValidator, QueryDiffCheckValidator

    def validate(params: dict = {}, view_options: dict = {}):
        QueryCheckValidator().validate({
            'name': 'test',
            'type': 'query',
            'params': params,
            'view_options': view_options,
        })

    def validate_diff(params: dict = {}, view_options: dict = {}):
        QueryDiffCheckValidator().validate({
            'name': 'test',
            'type': 'query_diff',
            'params': params,
            'view_options': view_options,
        })

    # query
    validate({
        'sql_template': 'select * from abc'
    })

    # diff in client
    validate_diff({
        'sql_template': 'select * from abc'
    })
    validate_diff({
        'sql_template': 'select * from abc',
        'base_sql_template': 'select * from abc',
    })

    # diff in warehouse
    validate_diff({
        'primary_keys': ['customer_id'],
        'sql_template': 'select * from abc',
    })
    validate_diff({
        'sql_template': 'select * from abc',
        'base_sql_template': 'select * from abc',
        'primary_keys': ['customer_id']
    })
    validate_diff({
        'sql_template': 'select * from abc',
        'base_sql_template': 'select * from abc',
        'primary_keys': ['customer_id']
    })

    # invalid
    with pytest.raises(ValueError):
        validate()

    with pytest.raises(ValueError):
        validate_diff()

    with pytest.raises(ValueError):
        validate_diff({'sql_template': 123, 'primary_keys': 'xyz'})

    with pytest.raises(ValueError):
        validate_diff({'sql_template': 's', 'primary_keys': 'xyz'})



from typing import Optional

from pydantic import BaseModel

from recce.core import default_context
from recce.models import Check
from recce.tasks import Task
from recce.tasks.core import TaskResultDiffer, CheckValidator
from recce.tasks.query import QueryMixin


class TopKDiffParams(BaseModel):
    model: str
    column_name: str
    k: Optional[int] = 10


class TopKDiffTask(Task, QueryMixin):
    def __init__(self, params):
        super().__init__()
        self.params = TopKDiffParams(**params)
        self.connection = None

    def _query_row_count_diff(self, dbt_adapter, base_relation, curr_relation, column):
        """
        Query the row count of the base and current relations

        :return: [base_total, base_valids, curr_total, curr_valids]
        """

        sql_template = r"""
        select count(*), count({{column}}) from {{ base_relation }}
        UNION ALL
        select count(*), count({{column}}) from {{ curr_relation }}
        """
        sql = dbt_adapter.generate_sql(sql_template, context=dict(
            base_relation=base_relation,
            curr_relation=curr_relation,
            column=column,
        ))
        _, table = dbt_adapter.execute(sql, fetch=True)

        result = (table[0][0], table[0][1], table[1][0], table[1][1])

        return (int(v) if v is not None else 0 for v in result)

    def _query_top_k(self, dbt_adapter, base_relation, curr_relation, column, k):
        sql_template = r"""
        WITH
        BASE_CAT as (
            select
                coalesce(cast({{column}} as {{ dbt.type_string() }}), '__null__') as category,
                count(*) as c
            from {{base_relation}}
            {% if not include_null %}
            where {{column}} is not null
            {% endif %}
            group by 1
        ),
        CURR_CAT as (
            select
                coalesce(cast({{column}} as {{ dbt.type_string() }}), '__null__') as category,
                count(*) as c
            from {{curr_relation}}
            {% if not include_null %}
            where {{column}} is not null
            {% endif %}
            group by 1
        )
        select
            coalesce(CURR_CAT.category, BASE_CAT.category) as category,
            coalesce(BASE_CAT.c, 0) as base_count,
            coalesce(CURR_CAT.c, 0) as curr_count
        from CURR_CAT
        full outer join BASE_CAT
        on CURR_CAT.category = BASE_CAT.category
        order by curr_count desc, base_count desc
        limit {{k}}
        """
        sql = dbt_adapter.generate_sql(sql_template, context=dict(
            base_relation=base_relation,
            curr_relation=curr_relation,
            column=column,
            k=k,
            include_null=False,
        ))
        _, table = dbt_adapter.execute(sql, fetch=True)

        categories = []
        base_counts = []
        curr_counts = []

        for row in table:
            categories.append(row[0] if row[0] != '__null__' else None)
            base_counts.append(int(row[1] if row[1] else 0))
            curr_counts.append(int(row[2] if row[2] else 0))

        return categories, base_counts, curr_counts

    def execute(self):

        from recce.adapter.dbt_adapter import DbtAdapter
        dbt_adapter: DbtAdapter = default_context().adapter

        with dbt_adapter.connection_named("query"):
            self.connection = dbt_adapter.get_thread_connection()
            model = self.params.model
            column = self.params.column_name
            k = self.params.k or 10

            base_relation = dbt_adapter.create_relation(model, base=True)
            if base_relation is None:
                raise ValueError(f"Model '{model}' not found in the manifest")

            curr_relation = dbt_adapter.create_relation(model, base=False)
            if curr_relation is None:
                raise ValueError(f"Model '{model}' not found in the manifest")

            self.check_cancel()
            categories, base_counts, curr_counts = self._query_top_k(
                dbt_adapter,
                base_relation,
                curr_relation,
                column,
                k
            )
            self.check_cancel()

            base_total, base_valids, curr_total, curr_valids = self._query_row_count_diff(
                dbt_adapter,
                base_relation,
                curr_relation,
                column
            )

            result = {
                'base': {
                    'values': categories,
                    'counts': base_counts,
                    'valids': base_valids,
                    'total': base_total,
                },
                'current': {
                    'values': categories,
                    'counts': curr_counts,
                    'valids': curr_valids,
                    'total': curr_total,
                },
            }
            return result

    def cancel(self):
        super().cancel()
        if self.connection:
            self.close_connection(self.connection)


class TopKDiffTaskResultDiffer(TaskResultDiffer):
    def _check_result_changed_fn(self, result):
        base = result.get('base')
        current = result.get('current')

        return TaskResultDiffer.diff(base, current)


class TopKDiffCheckValidator(CheckValidator):

    def validate_check(self, check: Check):
        try:
            TopKDiffParams(**check.params)
        except Exception as e:
            raise ValueError(f"Invalid check: {str(e)}")



import os
import textwrap
import uuid
from io import StringIO

from dbt.contracts.graph.nodes import ModelNode, SnapshotNode

from recce.adapter.dbt_adapter import DbtAdapter, as_manifest, load_manifest
from recce.core import RecceContext


class DbtTestHelper:

    def __init__(self):
        schema_prefix = "schema_" + uuid.uuid4().hex
        self.base_schema = f"{schema_prefix}_base"
        self.curr_schema = f"{schema_prefix}_curr"

        current_dir = os.path.dirname(os.path.abspath(__file__))
        project_dir = os.path.join(current_dir, 'test_proj')
        profiles_dir = project_dir
        manifest_path = os.path.join(project_dir, 'manifest.json')

        dbt_adapter = DbtAdapter.load(
            no_artifacts=True,
            project_dir=project_dir,
            profiles_dir=profiles_dir,
        )

        context = RecceContext()
        context.adapter_type = 'dbt'
        context.adapter = dbt_adapter
        context.schema_prefix = schema_prefix
        self.adapter = dbt_adapter
        self.context = context
        curr_writable_manifest = load_manifest(manifest_path)
        base_writable_manifest = load_manifest(manifest_path)

        self.curr_manifest = as_manifest(curr_writable_manifest)
        self.base_manifest = as_manifest(base_writable_manifest)
        self.context = context

        self.adapter.execute(f"CREATE schema IF NOT EXISTS {self.base_schema}")
        self.adapter.execute(f"CREATE schema IF NOT EXISTS {self.curr_schema}")
        self.adapter.set_artifacts(
            base_writable_manifest,
            curr_writable_manifest,
            self.curr_manifest,
            self.base_manifest
        )

    def create_model(
        self,
        model_name,
        base_csv=None,
        curr_csv=None,
        base_sql=None,
        curr_sql=None,
        depends_on=[],
        disabled=False,
        unique_id=None,
        package_name="recce_test",
    ):
        # unique_id = f"model.{package_name}.{model_name}"
        unique_id = unique_id if unique_id else model_name

        def _add_model_to_manifest(base, raw_code):
            if base:
                schema = self.base_schema
                manifest = self.base_manifest
            else:
                schema = self.curr_schema
                manifest = self.curr_manifest

            node = ModelNode.from_dict({
                "resource_type": "model",
                "name": model_name,
                "package_name": package_name,
                "path": "",
                "original_file_path": "",
                "unique_id": unique_id,
                "fqn": [
                    package_name,
                    model_name,
                ],
                "schema": schema,
                "alias": model_name,
                "checksum": {
                    "name": "sha256",
                    "checksum": hash(raw_code),
                },
                "raw_code": raw_code,
                "config": {
                    "materialized": "table",
                    "tags": ["test_tag"],
                },
                "tags": ["test_tag"],
                "depends_on": {
                    "nodes": depends_on
                },
            })

            if disabled:
                manifest.add_disabled_nofile(node)
            else:
                manifest.add_node_nofile(node)
            return node

        dbt_adapter = self.adapter
        with dbt_adapter.connection_named('create model'):
            import pandas as pd
            if base_csv:
                base_csv = textwrap.dedent(base_csv)
                _add_model_to_manifest(True, base_csv)
                df_base = pd.read_csv(StringIO(base_csv))
                dbt_adapter.execute(f"CREATE TABLE {self.base_schema}.{model_name} AS SELECT * FROM df_base")
            elif base_sql:
                bas_sql = textwrap.dedent(base_sql)
                _add_model_to_manifest(True, base_sql)

            if curr_csv:
                curr_csv = textwrap.dedent(curr_csv)
                _add_model_to_manifest(False, curr_csv)
                df_curr = pd.read_csv(StringIO(curr_csv))
                dbt_adapter.execute(f"CREATE TABLE {self.curr_schema}.{model_name} AS SELECT * FROM df_curr")
            elif curr_sql:
                curr_sql = textwrap.dedent(curr_sql)
                _add_model_to_manifest(False, curr_sql)

        self.adapter.set_artifacts(
            self.base_manifest.writable_manifest(),
            self.curr_manifest.writable_manifest(),
            self.curr_manifest,
            self.base_manifest)

    def remove_model(self, model_name):
        dbt_adapter = self.adapter
        with dbt_adapter.connection_named('cleanup'):
            dbt_adapter.execute(f"DROP TABLE IF EXISTS {self.base_schema}.{model_name}")
            dbt_adapter.execute(f"DROP TABLE IF EXISTS {self.curr_schema}.{model_name} ")

    def cleanup(self):
        dbt_adapter = self.adapter
        with dbt_adapter.connection_named('cleanup'):
            dbt_adapter.execute(f"DROP SCHEMA IF EXISTS {self.base_schema} CASCADE")
            dbt_adapter.execute(f"DROP SCHEMA IF EXISTS {self.curr_schema} CASCADE")

    def create_snapshot(self, sanpshot_name, base_csv, curr_csv, depends_on=[]):
        package_name = "recce_test"
        # unique_id = f"model.{package_name}.{model_name}"
        unique_id = sanpshot_name

        def _add_snapshot_to_manifest(base, raw_code):
            if base:
                schema = self.base_schema
                manifest = self.base_manifest
            else:
                schema = self.curr_schema
                manifest = self.curr_manifest

            node = SnapshotNode.from_dict({
                "resource_type": "snapshot",
                "name": sanpshot_name,
                "package_name": package_name,
                "path": "",
                "original_file_path": "",
                "unique_id": unique_id,
                "fqn": [
                    package_name,
                    sanpshot_name,
                ],
                "schema": schema,
                "alias": sanpshot_name,
                "checksum": {
                    "name": "sha256",
                    "checksum": hash(raw_code),
                },
                "raw_code": raw_code,
                "config": {
                    "materialized": "snapshot",
                    "tags": ["test_tag"],
                },
                "tags": ["test_tag"],
                "depends_on": {
                    "nodes": depends_on
                },
            })
            manifest.add_node_nofile(node)
            return node

        base_csv = textwrap.dedent(base_csv)
        curr_csv = textwrap.dedent(curr_csv)

        _add_snapshot_to_manifest(True, base_csv)
        _add_snapshot_to_manifest(False, curr_csv)

        import pandas as pd
        df_base = pd.read_csv(StringIO(base_csv))
        df_curr = pd.read_csv(StringIO(curr_csv))
        dbt_adapter = self.adapter
        with dbt_adapter.connection_named('create model'):
            dbt_adapter.execute(f"CREATE TABLE {self.base_schema}.{sanpshot_name} AS SELECT * FROM df_base")
            dbt_adapter.execute(f"CREATE TABLE {self.curr_schema}.{sanpshot_name} AS SELECT * FROM df_curr")
        self.adapter.set_artifacts(
            self.base_manifest.writable_manifest(),
            self.curr_manifest.writable_manifest(),
            self.curr_manifest,
            self.base_manifest)

    def remove_snapshot(self, snapshot_name):
        dbt_adapter = self.adapter
        with dbt_adapter.connection_named('cleanup'):
            dbt_adapter.execute(f"DROP TABLE IF EXISTS {self.base_schema}.{snapshot_name}")
            dbt_adapter.execute(f"DROP TABLE IF EXISTS {self.curr_schema}.{snapshot_name} ")



from typing import Optional, Literal

from pydantic import BaseModel

from recce.models import Check
from recce.tasks.core import CheckValidator


class LineageDiffParams(BaseModel):
    select: Optional[str] = None
    exclude: Optional[str] = None
    packages: Optional[list[str]] = None
    view_mode: Optional[Literal['all', 'changed_models']] = None


class LineageDiffCheckValidator(CheckValidator):
    def validate_check(self, check: Check):
        LineageDiffParams(**check.params)
        LineageDiffParams(**check.view_options)



import logging
from abc import ABC, abstractmethod
from typing import Optional, Callable, Set, Literal

from recce.models.types import LineageDiff, NodeDiff
from recce.state import ArtifactsRoot

# from dbt.contracts.graph.nodes import ManifestNode

logger = logging.getLogger('uvicorn')


class BaseAdapter(ABC):

    @classmethod
    def load(cls, **kwargs):
        raise NotImplementedError()

    @abstractmethod
    def get_lineage(self, base: Optional[bool] = False):
        raise NotImplementedError()

    def get_lineage_diff(self) -> LineageDiff:
        base = self.get_lineage(base=True)
        current = self.get_lineage(base=False)
        keys = {
            *base.get('nodes', {}).keys(),
            *current.get('nodes', {}).keys()
        }

        # for each node, compare the base and current lineage
        diff = {}
        for key in keys:
            base_node = base.get('nodes', {}).get(key)
            curr_node = current.get('nodes', {}).get(key)
            if base_node and curr_node:
                base_checksum = base_node.get('checksum', {}).get('checksum')
                curr_checksum = curr_node.get('checksum', {}).get('checksum')

                if base_checksum is not None and base_checksum == curr_checksum:
                    continue

                diff[key] = NodeDiff(change_status='modified', change_category='breaking')
            elif base_node:
                diff[key] = NodeDiff(chnage_status='removed')
            elif curr_node:
                diff[key] = NodeDiff(change_status='added')
        return LineageDiff(
            base=base,
            current=current,
            diff=diff,
        )

    @abstractmethod
    def select_nodes(
        self,
        select: Optional[str] = None,
        exclude: Optional[str] = None,
        packages: Optional[list[str]] = None,
        view_mode: Optional[Literal['all', 'changed_models']] = None,
    ) -> Set[str]:
        raise NotImplementedError()

    @abstractmethod
    def get_model(self, model_id: str, base=False):
        raise NotImplementedError()

    @abstractmethod
    def get_node_name_by_id(self, unique_id):
        raise NotImplementedError()

    @abstractmethod
    def support_tasks(self):
        """
        Get the adapter support tasks. Should be implemented by subclass.
        The tasks support map is a dictionary that maps all the Recce tasks are supported by the adapter or not.
        True means supported, False means not supported.
        """
        raise NotImplementedError()

    def start_monitor_artifacts(self, callback: Callable = None):
        pass

    def stop_monitor_artifacts(self):
        pass

    def start_monitor_base_env(self, callback: Callable = None):
        pass

    def stop_monitor_base_env(self):
        pass

    def refresh(self, refresh_file_path: str = None):
        pass

    def export_artifacts(self) -> ArtifactsRoot:
        return ArtifactsRoot(base={}, current={})

    def import_artifacts(self, artifacts: ArtifactsRoot, merge: bool = False):
        pass

    def find_node_by_name(self, node_name, base=False):
        logger.info("Deprecated method find_node_by_name. It returns dbt's type")
        manifest = self.curr_manifest if base is False else self.base_manifest
        for key, node in manifest.nodes.items():
            if node.name == node_name:
                return node
        return None

    def get_node_by_name(self, node_name):
        node = self.find_node_by_name(node_name) or self.find_node_by_name(node_name, base=True)
        return node



import pytest

from recce.tasks import ValueDiffTask, ValueDiffDetailTask


def test_value_diff(dbt_test_helper):
    csv_data_curr = """
        customer_id,name,age
        1,Alice,30
        2,Bob,25
        3,Charlie,35
        """

    csv_data_base = """
        customer_id,name,age
        1,Alice,35
        2,Bob,25
        3,Charlie,35
        """

    dbt_test_helper.create_model("customers", csv_data_base, csv_data_curr)
    params = dict(model='customers', primary_key=['customer_id'])
    task = ValueDiffTask(params)
    run_result = task.execute()
    assert len(run_result.data.columns) == 3
    assert len(run_result.data.data) == 3

    params = dict(model='customers', primary_key=['customer_id'])
    task = ValueDiffDetailTask(params)
    run_result = task.execute()
    assert len(run_result.columns) == 5
    assert len(run_result.data) == 2


def test_validator():
    from recce.tasks.valuediff import ValueDiffCheckValidator

    def validate(params: dict = {}, view_options: dict = {}):
        ValueDiffCheckValidator().validate({
            'name': 'test',
            'type': 'value_diff',
            'params': params,
            'view_options': view_options,
        })

    validate({
        'model': 'customers',
        'primary_key': 'customer_id',
    })
    validate({
        'model': 'customers',
        'primary_key': ['customer_id'],
    })
    validate({
        'model': 'customers',
        'primary_key': ['customer_id'],
        'columns': ['name', 'age'],
    })

    with pytest.raises(ValueError):
        validate({
        })

    with pytest.raises(ValueError):
        validate({
            'model': 'customers',
        })

    with pytest.raises(ValueError):
        validate({
            'model': 'customers',
            'primary_key': ['customer_id'],
            'columns': 'name',
        })



from sqlglot.optimizer import optimize


def is_breaking_change(original_sql, modified_sql, dialect=None):
    if original_sql == modified_sql:
        return False

    try:
        from sqlglot import parse_one, diff
        from sqlglot.diff import Insert, Keep
        import sqlglot.expressions as exp
    except ImportError:
        return True

    try:
        original_ast = optimize(parse_one(original_sql, dialect=dialect))
        modified_ast = optimize(parse_one(modified_sql, dialect=dialect))
    except Exception:
        return True

    if not isinstance(original_ast, exp.Select) or not isinstance(modified_ast, exp.Select):
        raise ValueError("Currently only SELECT statements are supported for comparison")

    edits = diff(original_ast, modified_ast, delta_only=True)

    inserted_expressions = {
        e.expression for e in edits if isinstance(e, Insert)
    }

    for edit in edits:
        if isinstance(edit, Insert):
            inserted_expr = edit.expression
            VALID_EXPRESSIONS = (
                exp.Where,
                exp.Join,
                exp.Order,
                exp.Group,
                exp.Having,
                exp.Limit,
                exp.Offset,
                exp.Window,
                exp.Union,
                exp.Intersect,
                exp.Except,
                exp.Merge,
                exp.Delete,
                exp.Update,
                exp.Insert,
                exp.Subquery,
            )

            if isinstance(inserted_expr, VALID_EXPRESSIONS):
                return True

            if (
                not isinstance(inserted_expr.parent, exp.Select) and
                inserted_expr.parent not in inserted_expressions
            ):
                return True
        elif not isinstance(edit, Keep):
            return True

    return False



from .core import Task
from .histogram import HistogramDiffTask
from .profile import ProfileDiffTask
from .query import QueryTask, QueryDiffTask, QueryBaseTask
from .rowcount import RowCountDiffTask
from .top_k import TopKDiffTask
from .valuediff import ValueDiffTask, ValueDiffDetailTask



# noinspection PyUnresolvedReferences
from tests.adapter.dbt_adapter.conftest import dbt_test_helper


def test_lineage_diff(dbt_test_helper):
    sql_model1 = """
    select a from T
    """

    sql_model2 = """
    select a from {{ ref("model1") }}
    """

    sql_model2_ = """
    select
    a,b
    from
    {{ ref("model1") }}
    """

    dbt_test_helper.create_model("model1", sql_model1, sql_model1)
    dbt_test_helper.create_model("model2", sql_model2, sql_model2_)
    result = dbt_test_helper.context.get_lineage_diff()
    nodediff = result.diff.get('model1')
    assert nodediff is None
    nodediff2 = result.diff.get('model2')
    assert nodediff2 is not None and nodediff2.change_status == 'modified' and nodediff2.change_category == 'non-breaking'



from unittest import TestCase
from unittest.mock import patch, MagicMock

from click.testing import CliRunner

from recce.cli import server as cli_command_server, run as cli_command_run
from recce.core import RecceContext
from recce.state import RecceStateLoader


def test_cmd_version():
    from recce.cli import version
    from recce import __version__
    runner = CliRunner()
    result = runner.invoke(version, [])
    assert result.exit_code == 0
    assert result.output.replace('\n', '') == __version__


class TestCommandServer(TestCase):
    def setUp(self):
        self.runner = CliRunner()
        pass

    @patch.object(RecceContext, 'verify_required_artifacts')
    @patch('recce.cli.uvicorn.run')
    def test_cmd_server(self, mock_run, mock_verify_required_artifacts):
        from recce.server import app
        mock_verify_required_artifacts.return_value = True, None
        self.runner.invoke(cli_command_server, ['--host', 'unittest', '--port', 5566])
        mock_run.assert_called_once_with(app, host='unittest', port=5566, lifespan='on')

    @patch('recce.cli.uvicorn.run')
    def test_cmd_server_with_cloud_without_password(self, mock_run):
        # Should fail if no password is provided
        result = self.runner.invoke(cli_command_server, ['--cloud'])
        assert result.exit_code == 1

    @patch('recce.cli.uvicorn.run')
    def test_cmd_server_with_cloud_without_token(self, mock_run):
        # Should fail if no token is provided
        result = self.runner.invoke(cli_command_server, ['--cloud', '--password', 'unittest'])
        assert result.exit_code == 1

    @patch.object(RecceContext, 'verify_required_artifacts')
    @patch('recce.util.recce_cloud.get_recce_cloud_onboarding_state')
    @patch('recce.cli.uvicorn.run')
    @patch('recce.cli.RecceStateLoader')
    def test_cmd_server_with_cloud(self, mock_state_loader_class, mock_run, mock_get_recce_cloud_onboarding_state,
                                   mock_verify_required_artifacts):
        mock_state_loader = MagicMock(spec=RecceStateLoader)
        mock_state_loader.verify.return_value = True
        mock_state_loader.review_mode = True
        mock_get_recce_cloud_onboarding_state.return_value = 'completed'
        mock_verify_required_artifacts.return_value = True, None

        mock_state_loader_class.return_value = mock_state_loader
        self.runner.invoke(cli_command_server, ['--cloud', '--password', 'unittest', '--cloud-token', 'unittest'])
        mock_state_loader_class.assert_called_once()
        mock_run.assert_called_once()

    @patch.object(RecceContext, 'verify_required_artifacts')
    @patch('os.path.isdir', side_effect=lambda path: True if path == 'existed_folder' else False)
    @patch('recce.cli.uvicorn.run')
    @patch('recce.server.AppState')
    def test_cmd_server_with_target_folder_only(self,
                                                mock_app_state, mock_run, mock_isdir, mock_verify_required_artifacts):
        mock_verify_required_artifacts.return_value = True, None
        self.runner.invoke(cli_command_server,
                           [
                               '--target-path', 'existed_folder',
                               '--target-base-path', 'non_existed_folder',
                           ])
        mock_run.assert_called_once()

        # Onboarding mode should be set to True
        app_state_call_args = mock_app_state.call_args
        app_state_flag = app_state_call_args.kwargs['flag']
        assert 'single_env_onboarding' in app_state_flag
        assert app_state_flag['single_env_onboarding'] is True
        assert 'show_relaunch_hint' in app_state_flag
        assert app_state_flag['show_relaunch_hint'] is True

        # The target_base_path should be set to the same as target_path
        verify_required_artifacts_args = mock_verify_required_artifacts.call_args
        assert verify_required_artifacts_args.kwargs['target_path'] == verify_required_artifacts_args.kwargs[
            'target_base_path']


class TestCommandRun(TestCase):
    def setUp(self):
        self.runner = CliRunner()
        pass

    @patch.object(RecceContext, 'verify_required_artifacts')
    @patch('recce.cli.cli_run')
    def test_cmd_run(self, mock_cli_run, mock_verify_required_artifacts):
        mock_verify_required_artifacts.return_value = True, None

        self.runner.invoke(cli_command_run, [])
        mock_cli_run.assert_called_once()



import hashlib
import os
import re
import sys
import threading
import uuid
from datetime import datetime, timezone
from hashlib import sha256

import sentry_sdk

from recce import is_ci_env, get_version, get_runner
from recce import yaml as pyml
from recce.event.collector import Collector
from recce.git import current_branch, hosting_repo
from recce.github import is_github_codespace, get_github_codespace_info, get_github_codespace_name, \
    get_github_codespace_available_at

USER_HOME = os.path.expanduser('~')
RECCE_USER_HOME = os.path.join(USER_HOME, '.recce')
RECCE_USER_PROFILE = os.path.join(RECCE_USER_HOME, 'profile.yml')
RECCE_USER_EVENT_PATH = os.path.join(RECCE_USER_HOME, '.unsend_events.json')

__version__ = get_version()
_collector = Collector()
user_profile_lock = threading.Lock()


def init():
    api_key = _get_api_key()
    user_profile = load_user_profile()

    # Amplitude init
    _collector.set_api_key(api_key)
    _collector.set_user_id(user_profile.get('user_id'))
    _collector.set_unsend_events_file(RECCE_USER_EVENT_PATH)

    # Sentry init
    sentry_env = _get_sentry_env()
    sentry_dns = _get_sentry_dns()
    release_version = __version__ if sentry_env != 'development' else None
    sentry_sdk.init(
        dsn=sentry_dns,
        environment=sentry_env,
        release=release_version,
        # Set traces_sample_rate to 1.0 to capture 100%
        # of transactions for performance monitoring.
        # We recommend adjusting this value in production.
        traces_sample_rate=1.0
    )
    sentry_sdk.set_tag('recce.version', __version__)
    sentry_sdk.set_tag('platform', sys.platform)
    sentry_sdk.set_tag('is_ci_env', is_ci_env())
    sentry_sdk.set_tag('is_github_codespace', is_github_codespace())
    sentry_sdk.set_tag('system_timezone', get_system_timezone())


def get_user_id():
    return load_user_profile().get('user_id')


def is_anonymous_tracking():
    return load_user_profile().get('anonymous_tracking', False)


def _get_sentry_dns():
    dns_file = os.path.normpath(os.path.join(os.path.dirname(__file__), 'SENTRY_DNS'))
    with open(dns_file) as f:
        dns = f.read().strip()
        return dns


def _get_sentry_env():
    if '.dev' in __version__:
        return 'development'
    elif re.match(r'^\d+\.\d+\.\d+\.\d{8}[a|b|rc]?.*$', __version__):
        return 'nightly'
    elif 'a' in __version__:
        return 'alpha'
    elif 'b' in __version__:
        return 'beta'
    elif 'rc' in __version__:
        return 'release-candidate'
    return 'production'


def _get_api_key():
    if os.getenv('RECCE_EVENT_API_KEY'):
        # For local testing purpose
        return os.getenv('RECCE_EVENT_API_KEY')

    config_file = os.path.abspath(os.path.join(os.path.dirname(__file__), 'CONFIG'))
    try:
        with open(config_file) as fh:
            config = pyml.load(fh)
            return config.get('event_api_key')
    except Exception:
        return None


def _generate_user_profile():
    try:
        os.makedirs(RECCE_USER_HOME, exist_ok=True)
    except Exception:
        # TODO: should show warning message but not raise exception
        print('Please disable command tracking to continue.')
        exit(1)
    if is_github_codespace() is True:
        salted_name = f'codespace-{get_github_codespace_name()}'
        user_id = hashlib.sha256(salted_name.encode()).hexdigest()
    else:
        user_id = uuid.uuid4().hex
    with open(RECCE_USER_PROFILE, 'w+') as f:
        pyml.dump({'user_id': user_id, 'anonymous_tracking': True}, f)
    return dict(user_id=user_id, anonymous_tracking=True)


def load_user_profile():
    with user_profile_lock:
        if not os.path.exists(RECCE_USER_PROFILE):
            user_profile = _generate_user_profile()
        else:
            with open(RECCE_USER_PROFILE, 'r') as f:
                user_profile = pyml.load(f)
                if user_profile.get('user_id') is None:
                    user_profile = _generate_user_profile()

        return user_profile


def update_user_profile(update_values):
    original = load_user_profile()
    original.update(update_values)
    with open(RECCE_USER_PROFILE, 'w+') as f:
        pyml.dump(original, f)
    return original


def flush_events(command=None):
    _collector.send_events()


def should_log_event():
    with open(RECCE_USER_PROFILE, 'r') as f:
        user_profile = pyml.load(f)
    # TODO: default anonymous_tracking to false if field is not present
    tracking = user_profile.get('anonymous_tracking', False)
    tracking = tracking and isinstance(tracking, bool)
    if not tracking:
        return False

    if not _collector.is_ready():
        return False

    return True


def log_event(prop, event_type, **kwargs):
    if should_log_event() is False:
        return

    repo = hosting_repo()
    if repo is not None:
        prop['repository'] = sha256(repo.encode()).hexdigest()

    branch = current_branch()
    if branch is not None:
        prop['branch'] = sha256(branch.encode()).hexdigest()

    runner = get_runner()
    if runner is not None:
        prop['runner_type'] = runner

        if runner == 'github codespaces':
            prop['codespaces_name'] = get_github_codespace_name()

    payload = dict(
        **prop,
    )

    _collector.log_event(payload, event_type)


def log_api_event(endpoint_name, prop):
    prop = dict(
        **prop,
        endpoint_name=endpoint_name,
    )
    log_event(prop, 'api_event')
    _collector.schedule_flush()


def log_load_state(command='server'):
    from recce.models import CheckDAO

    checks = 0
    preset_checks = 0

    for check in CheckDAO().list():
        checks += 1
        if check.is_preset:
            preset_checks += 1

    prop = dict(
        command=command,
        checks=checks,
        preset_checks=preset_checks,
    )

    log_event(prop, 'load_state')
    if command == 'server':
        _collector.schedule_flush()


def log_codespaces_events(command):
    # Only log when the recce is running in GitHub Codespaces
    codespace = get_github_codespace_info()
    if codespace is None:
        return

    user_prop = dict(
        location=codespace.get('location'),
        is_prebuild=codespace.get('prebuild', False),
    )

    prop = dict(
        machine=codespace.get('machine', {}).get('display_name'),
        codespaces_name=get_github_codespace_name(),
    )

    # Codespace created event, send once
    codespace_created_at = load_user_profile().get('codespace_created_at')
    if codespace_created_at is None:
        created_at = datetime.fromisoformat(codespace.get('created_at'))
        prop['state'] = 'created'
        _collector.log_event(prop, 'codespace_instance', event_triggered_at=created_at,
                             user_properties=user_prop)
        update_user_profile({'codespace_created_at': codespace.get('created_at')})

    # Codespace available event, send multiple times as start/stop it
    available_at = get_github_codespace_available_at(codespace)
    if available_at and available_at.isoformat() != load_user_profile().get('codespace_available_at'):
        prop['state'] = 'available'
        _collector.log_event(prop, 'codespace_instance', event_triggered_at=available_at,
                             user_properties=user_prop)
        update_user_profile({'codespace_available_at': available_at.isoformat()})

    # Codespace instance event should be flushed immediately
    _collector.send_events()


def log_single_env_event():
    prop = dict(
        action='launch_server',
    )
    log_event(prop, '[Experiment] single_environment')
    _collector.schedule_flush()


def capture_exception(e):
    user_id = load_user_profile().get('user_id')
    if is_ci_env() is True:
        user_id = f"{user_id}_CI"

    sentry_sdk.set_tag("user_id", user_id)
    sentry_sdk.capture_exception(e)


def flush_exceptions():
    sentry_sdk.flush()


def set_exception_tag(key, value):
    sentry_sdk.set_tag(key, value)


def get_system_timezone():
    return datetime.now(timezone.utc).astimezone().tzinfo



import gzip
import json
import os
import shutil
import tarfile
import tempfile
from urllib.parse import urlencode

import requests
from rich.console import Console

from recce.git import hosting_repo, commit_hash_from_branch, current_branch
from recce.state import s3_sse_c_headers
from recce.util.recce_cloud import RecceCloud, PresignedUrlMethod


def verify_artifacts_path(target_path: str) -> bool:
    """
    Verify if the target path is a valid artifacts path.

    :param target_path: the target path to check
    :return: True if the target path is a valid artifacts path, False otherwise
    """
    if not target_path:
        return False

    if not os.path.exists(target_path):
        return False

    if not os.path.isdir(target_path):
        return False

    required_artifacts_files = [
        'manifest.json',
        'catalog.json'
    ]

    if all(f in os.listdir(target_path) for f in required_artifacts_files):
        # Check if the required files are present in the target path
        return True

    return False


def parse_dbt_version(file_path: str) -> str:
    with open(file_path, 'r') as f:
        data = json.load(f)

    dbt_version = data.get('metadata', {}).get('dbt_version', None)
    return dbt_version


def archive_artifacts(target_path: str) -> (str, str):
    if verify_artifacts_path(target_path) is False:
        raise Exception(f'Invalid target path: {target_path}')

    manifest_path = os.path.join(target_path, 'manifest.json')
    catalog_path = os.path.join(target_path, 'catalog.json')

    dbt_version = parse_dbt_version(manifest_path)
    if dbt_version is None:
        raise Exception('Failed to parse dbt version from manifest.json')

    # prepare the temporary artifacts path
    tmp_dir = tempfile.mkdtemp()
    artifacts_tar_path = os.path.join(tmp_dir, 'dbt_artifacts.tar')
    artifacts_tar_gz_path = artifacts_tar_path + '.gz'

    with tarfile.open(artifacts_tar_path, 'w') as tar:
        tar.add(manifest_path, arcname='manifest.json')
        tar.add(catalog_path, arcname='catalog.json')

    # Compress the tar file
    with open(artifacts_tar_path, 'rb') as f_in, gzip.open(artifacts_tar_gz_path, 'wb') as f_out:
        f_out.writelines(f_in)

    # Clean up the temporary directory
    try:
        os.remove(artifacts_tar_path)
    except FileNotFoundError:
        pass

    return artifacts_tar_gz_path, dbt_version


def upload_dbt_artifacts(target_path: str, branch: str, token: str, password: str, debug: bool = False):
    console = Console()
    if verify_artifacts_path(target_path) is False:
        console.print(f"[[red]Error[/red]] Invalid target path: {target_path}")
        console.print("Please provide a valid target path containing manifest.json and catalog.json.")
        return 1

    if branch != current_branch():
        console.print(
            f"[[yellow]Warning[/yellow]] You are uploading the dbt artifacts as branch '{branch}'. "
            f"However, the current branch is '{current_branch()}'."
        )
        console.print("Please make sure you are uploading the dbt artifacts to the correct branch.")

    compress_file_path, dbt_version = archive_artifacts(target_path)
    repo = hosting_repo()
    sha = commit_hash_from_branch(branch)
    metadata = {
        'commit': sha,
        'dbt_version': dbt_version
    }

    # Get the presigned URL for uploading the artifacts
    presigned_url = RecceCloud(token).get_presigned_url(
        method=PresignedUrlMethod.UPLOAD,
        repository=repo,
        artifact_name='dbt_artifacts.tar.gz',
        branch=branch,
        metadata=metadata
    )

    if debug:
        console.rule('Debug information', style='blue')
        console.print(f'Branch: {branch}')
        console.print(f'Commit hash: {sha}')
        console.print(f'GitHub repository: {repo}')
        console.print(f'Artifact path: {compress_file_path}')
        console.print(f'DBT version: {dbt_version}')
        console.print(f'Presigned URL: {presigned_url}')
    console.print(f'Uploading the dbt artifacts from path "{target_path}" to branch "{branch}"')

    # Upload the compressed artifacts

    headers = s3_sse_c_headers(password)
    if metadata:
        headers['x-amz-tagging'] = urlencode(metadata)
    response = requests.put(presigned_url, data=open(compress_file_path, 'rb').read(), headers=headers)
    if response.status_code != 200:
        raise Exception({response.text})

    # Clean up the compressed artifacts
    try:
        # Remove the compressed artifacts
        os.remove(compress_file_path)
        # Clean up the temporary directory
        os.rmdir(os.path.dirname(compress_file_path))
    except FileNotFoundError:
        pass


def download_dbt_artifacts(target_path: str, branch: str, token: str, password: str,
                           force: bool = False,
                           debug: bool = False):
    console = Console()
    repo = hosting_repo()
    sha = None
    dbt_version = None

    presigned_url, tags = RecceCloud(token).get_download_presigned_url_with_tags(
        repository=repo,
        artifact_name='dbt_artifacts.tar.gz',
        branch=branch,
    )
    if tags:
        sha = tags.get('commit')
        dbt_version = tags.get('dbt_version')

    if debug:
        console.rule('Debug information', style='blue')
        console.print(f'Git Branch: {branch}')
        console.print(f'Git Commit: {sha}')
        console.print(f'GitHub repository: {repo}')
        console.print(f'DBT version: {dbt_version}')
    console.print(f'Downloading from branch: "{branch}" and extracting to "{target_path}"')

    headers = s3_sse_c_headers(password)
    response = requests.get(presigned_url, headers=headers)

    if response.status_code != 200:
        raise Exception(response.text)

    if os.path.exists(target_path):
        if not force:
            raise Exception(
                f'Path {target_path} already exists. Please provide a new path or use \'--force\' option to overwrite the existing folder.')
        console.print(f'[[yellow]Warning[/yellow]] Overwrite existing path: {target_path}')
        shutil.rmtree(target_path)
    os.mkdir(target_path)

    tar_gz_file = os.path.join(target_path, 'dbt_artifacts.tar.gz')
    with open(tar_gz_file, 'wb') as f:
        f.write(response.content)

    with tarfile.open(tar_gz_file, 'r') as tar:
        tar.extractall(path=target_path)

    # Clean up the compressed artifacts
    try:
        # Remove the compressed artifacts
        os.remove(tar_gz_file)
    except FileNotFoundError:
        pass
    return 0



import pytest

from recce.tasks import RowCountDiffTask


def test_row_count(dbt_test_helper):
    csv_data_curr = """
        customer_id,name,age
        1,Alice,30
        2,Bob,25
        3,Charlie,35
        """

    csv_data_base = """
        customer_id,name,age
        1,Alice,35
        2,Bob,25
        """

    dbt_test_helper.create_model("customers", csv_data_base, csv_data_curr, unique_id='model.customers')
    task = RowCountDiffTask(dict(node_names=['customers']))
    run_result = task.execute()
    assert run_result['customers']['base'] == 2
    assert run_result['customers']['curr'] == 3

    task = RowCountDiffTask(dict(node_names=['customers_']))
    run_result = task.execute()
    assert run_result['customers_']['base'] is None
    assert run_result['customers_']['curr'] is None

    task = RowCountDiffTask(dict(node_ids=['model.customers']))
    run_result = task.execute()
    assert run_result['customers']['base'] == 2
    assert run_result['customers']['curr'] == 3


def test_row_count_with_selector(dbt_test_helper):
    csv_data_1 = """
        customer_id,name,age
        1,Alice,30
        2,Bob,25
        3,Charlie,35
        """

    csv_data_2 = """
        customer_id,name,age
        1,Alice,35
        2,Bob,25
        """

    dbt_test_helper.create_model("model_1", csv_data_1, csv_data_2, depends_on=[])
    dbt_test_helper.create_model("model_2", csv_data_1, csv_data_1, depends_on=['model_1'],
                                 package_name='other_package')
    task = RowCountDiffTask(dict(select='model_1'))
    run_result = task.execute()
    assert len(run_result) == 1

    task = RowCountDiffTask(dict(select='model_1+'))
    run_result = task.execute()
    assert len(run_result) == 2


def test_validator():
    from recce.tasks.rowcount import RowCountDiffCheckValidator

    validator = RowCountDiffCheckValidator()

    def validate(params: dict):
        validator.validate({
            'name': 'test',
            'type': 'row_count_diff',
            'params': params,
        })

    # Select all modesl
    validate({})

    # Select by node name
    validate({
        'node_names': ['abc'],
    })
    with pytest.raises(ValueError):
        validate({
            'node_names': 'abc',
        })

    # Select by node id
    validate({
        'node_ids': ['model.abc'],
    })

    # Select by selector
    validate({
        'select': 'customers',
        'exclude': 'customers',
        'packages': ['jaffle_shop'],
        'view_mode': 'all',
    })

    # packages should be an array
    with pytest.raises(ValueError):
        validate({
            'packages': 'jaffle_shop',
        })

    # view_mode should be 'all' or 'changed_models'
    validate({
        'view_mode': None,
    })
    validate({
        'view_mode': 'all',
    })
    with pytest.raises(ValueError):
        validate({
            'view_mode': 'abc',
        })



from typing import TypedDict, Optional, List, Union

from pydantic import BaseModel

from .core import Task, TaskResultDiffer, CheckValidator
from .dataframe import DataFrame
from ..core import default_context
from ..exceptions import RecceException
from ..models import Check


class ValueDiffParams(BaseModel):
    model: str
    primary_key: Union[str, List[str]]
    columns: Optional[List[str]] = None


class ValueDiffResult(BaseModel):
    class Summary(BaseModel):
        total: int
        added: int
        removed: int

    summary: Summary
    data: DataFrame


class ValueDiffMixin:
    def _verify_dbt_packages_deps(self, dbt_adapter):
        for macro_name, macro in dbt_adapter.manifest.macros.items():
            if macro.package_name == 'audit_helper':
                break
        else:
            raise RecceException(
                r"Package 'audit_helper' not found. Please refer to the link to install: https://hub.getdbt.com/dbt-labs/audit_helper/")

        for macro_name, macro in dbt_adapter.manifest.macros.items():
            if macro.package_name == 'dbt_utils' and macro.name == 'generate_surrogate_key':
                self.legacy_surrogate_key = False
                break

    def _verify_primary_key(self, dbt_adapter, primary_key: Union[str, List[str]], model: str):
        self.update_progress(message=f"Verify primary key: {primary_key}")
        composite = True if isinstance(primary_key, List) else False

        if composite:
            if len(primary_key) == 0:
                raise RecceException("Primary key cannot be empty")
            sql_template = r"""{{ adapter.dispatch('test_unique_combination_of_columns', 'dbt_utils')(relation, primary_key) }}"""
        else:
            if primary_key is None or len(primary_key) == 0:
                raise RecceException("Primary key cannot be empty")
            sql_template = r"""{{ adapter.dispatch('test_unique', 'dbt')(relation, primary_key) }}"""

        # check primary keys
        for base in [True, False]:

            relation = dbt_adapter.create_relation(model, base)
            context = dict(
                relation=relation,
                primary_key=primary_key,
            )

            sql = dbt_adapter.generate_sql(sql_template, context=context)
            sql_test = f"""SELECT COUNT(*) AS INVALIDS FROM ({sql}) AS T"""

            response, table = dbt_adapter.adapter.execute(sql_test, fetch=True)
            for row in table.rows:
                invalids = row[0]
                if invalids > 0:
                    raise RecceException(
                        f"Invalid primary key: {primary_key}. The column should be unique. Please check by this sql: '{sql}'")
                break
            else:
                # it will never happen unless we use a wrong check sql
                raise RecceException('Cannot verify primary key')


class ValueDiffTask(Task, ValueDiffMixin):

    def __init__(self, params):
        super().__init__()
        self.params = ValueDiffParams(**params)
        self.connection = None
        self.legacy_surrogate_key = True

    def _query_value_diff(self, dbt_adpter, primary_key: Union[str, List[str]], model: str,
                          columns: List[str] = None):
        import agate

        column_groups = {}
        composite = True if isinstance(primary_key, List) else False

        if columns is None or len(columns) == 0:
            base_columns = [column.column for column in dbt_adpter.get_columns(model, base=True)]
            curr_columns = [column.column for column in dbt_adpter.get_columns(model, base=False)]
            columns = [column for column in base_columns if column in curr_columns]
        completed = 0

        if composite:
            for primary_key_comp in primary_key[::-1]:
                if primary_key_comp not in columns:
                    columns.insert(0, primary_key_comp)
        else:
            if primary_key not in columns:
                columns.insert(0, primary_key)

        sql_template = r"""
        {% set a_query %}
            select {{ __PRIMARY_KEY__ }} as _pk, * from {{ base_relation }}
        {% endset %}

        {% set b_query %}
            select {{ __PRIMARY_KEY__ }} as _pk, * from {{ curr_relation }}
        {% endset %}

        {{ audit_helper.compare_column_values(
            a_query=a_query,
            b_query=b_query,
            primary_key="_pk",
            column_to_compare=column_to_compare
        ) }}
        """

        if composite:
            if self.legacy_surrogate_key:
                new_primary_key = 'dbt_utils.surrogate_key(primary_key)'
            else:
                new_primary_key = 'dbt_utils.generate_surrogate_key(primary_key)'
        else:
            new_primary_key = 'primary_key'
        sql_template = sql_template.replace('__PRIMARY_KEY__', new_primary_key)

        for column in columns:
            self.update_progress(message=f"Diff column: {column}", percentage=completed / len(columns))

            sql = dbt_adpter.generate_sql(sql_template, context=dict(
                base_relation=dbt_adpter.create_relation(model, base=True),
                curr_relation=dbt_adpter.create_relation(model, base=False),
                primary_key=primary_key,
                column_to_compare=column,
            ))

            _, table = dbt_adpter.execute(sql, fetch=True)
            for row in table.rows:
                # data example:
                # ('COLUMN_NAME', 'MATCH_STATUS', 'COUNT_RECORDS', 'PERCENT_OF_TOTAL')
                # ('EVENT_ID', '✅: perfect match', 158601510, Decimal('100.00'))
                column_name, column_state, row_count, total_rate = row
                if 'column_name' == row[0].lower():
                    # skip column names
                    return

                #
                # sample data like this:
                #     https://github.com/dbt-labs/dbt-audit-helper/blob/main/macros/compare_column_values.sql
                #
                #     '✅: perfect match'            -> matched
                #     '✅: both are null'            -> matched
                #     '🤷: missing from a'           -> row added
                #     '🤷: missing from b'           -> row removed
                #     '🤷: value is null in a only'  -> mismatched
                #     '🤷: value is null in b only'  -> mismatched
                #     '🙅: values do not match'      -> mismatched
                #     'unknown'                      -> this should never happen
                # end as match_status,

                if column_name not in column_groups:
                    column_groups[column_name] = dict(added=0, removed=0, mismatched=0, matched=0)
                if 'perfect match' in column_state:
                    column_groups[column_name]['matched'] += row_count
                if 'both are null' in column_state:
                    column_groups[column_name]['matched'] += row_count
                if 'missing from a' in column_state:
                    column_groups[column_name]['added'] += row_count
                if 'missing from b' in column_state:
                    column_groups[column_name]['removed'] += row_count
                if 'value is null in a only' in column_state:
                    column_groups[column_name]['mismatched'] += row_count
                if 'value is null in b only' in column_state:
                    column_groups[column_name]['mismatched'] += row_count
                if 'values do not match' in column_state:
                    column_groups[column_name]['mismatched'] += row_count

            # Cancel as early as possible
            self.check_cancel()

            completed = completed + 1

        first = list(column_groups.values())[0]
        added = first['added']
        removed = first['removed']
        common = first['matched'] + first['mismatched']
        total = common + added + removed

        row = []
        for k, v in column_groups.items():
            if composite and k.lower() == "_pk":
                continue
            # This is incorrect when there are one side null
            # https://github.com/dbt-labs/dbt-audit-helper/blob/main/macros/compare_column_values.sql#L20-L23
            # matched = v['matched']
            matched = common - v['mismatched']
            rate = None if common == 0 else matched / common
            record = [k, matched, rate]
            row.append(record)

        column_names = ['column', 'matched', 'matched_p']
        column_types = [agate.Text(), agate.Number(), agate.Number()]
        table = agate.Table(row, column_names=column_names, column_types=column_types)

        return ValueDiffResult(
            summary=ValueDiffResult.Summary(total=total, added=added, removed=removed),
            data=DataFrame.from_agate(table),
        )

    def execute(self):
        dbt_adapter = default_context().adapter

        with dbt_adapter.connection_named("value diff"):
            self.connection = dbt_adapter.get_thread_connection()

            primary_key: Union[str, List[str]] = self.params.primary_key
            model: str = self.params.model
            columns: List[str] = self.params.columns

            self._verify_dbt_packages_deps(dbt_adapter)
            self.check_cancel()

            self._verify_primary_key(dbt_adapter, primary_key, model)
            self.check_cancel()

            return self._query_value_diff(dbt_adapter, primary_key, model, columns=columns)

    def cancel(self):
        super().cancel()
        from recce.adapter.dbt_adapter import DbtAdapter

        if self.connection:
            adapter: DbtAdapter = default_context().adapter
            with adapter.connection_named("cancel"):
                adapter.cancel(self.connection)


class ValueDiffTaskResultDiffer(TaskResultDiffer):

    def _check_result_changed_fn(self, result):
        is_changed = False
        summary = result.get('summary', {})
        added = summary.get('added', 0)
        removed = summary.get('removed', 0)
        changes = {
            'column_changed': []
        }

        if added > 0:
            is_changed = True
            changes['row_added'] = added

        if removed > 0:
            is_changed = True
            changes['row_removed'] = removed

        row_data = result.get('data', {}).get('data', [])
        for row in row_data:
            column, matched, matched_p = row
            if float(matched_p) < 1.0:
                # if there is any mismatched, we consider it as changed
                is_changed = True
                changes['column_changed'].append({
                    'column': column,
                    'matched': matched,
                    'matched_p': matched_p,
                })

        return changes if is_changed else None


class ValueDiffDetailParams(TypedDict):
    primary_key: str
    model: str
    columns: List[str]


class ValueDiffDetailResult(DataFrame):
    pass


class ValueDiffDetailTask(Task, ValueDiffMixin):

    def __init__(self, params):
        super().__init__()
        self.params = ValueDiffParams(**params)
        self.connection = None
        self.legacy_surrogate_key = True

    def _query_value_diff(self, dbt_adapter, primary_key: Union[str, List[str]], model: str, columns: List[str] = None):

        composite = True if isinstance(primary_key, List) else False

        if columns is None or len(columns) == 0:
            base_columns = [column.column for column in dbt_adapter.get_columns(model, base=True)]
            curr_columns = [column.column for column in dbt_adapter.get_columns(model, base=False)]
            columns = [column for column in base_columns if column in curr_columns]

        if composite:
            for primary_key_comp in primary_key[::-1]:
                if primary_key_comp not in columns:
                    columns.insert(0, primary_key_comp)
        else:
            if primary_key not in columns:
                columns.insert(0, primary_key)

        sql_template = r"""
        {% set col_list %}
            {%- for col in columns %}
                {{ col|trim }}
                {%- if not loop.last %},{{ '\n  ' }}{%- endif -%}
            {%- endfor -%}
        {% endset %}

        {% set a_query %}
            select {{col_list}} from {{ base_relation }}
        {% endset %}

        {% set b_query %}
            select {{col_list}} from {{ curr_relation }}
        {% endset %}

        {{ audit_helper.compare_queries(
            a_query=a_query,
            b_query=b_query,
            primary_key=__PRIMARY_KEY__,
            summarize=False,
        ) }} limit {{ limit }}
        """

        if composite:
            if self.legacy_surrogate_key:
                new_primary_key = 'dbt_utils.surrogate_key(primary_key)'
            else:
                new_primary_key = 'dbt_utils.generate_surrogate_key(primary_key)'
        else:
            new_primary_key = 'primary_key'
        sql_template = sql_template.replace('__PRIMARY_KEY__', new_primary_key)

        sql = dbt_adapter.generate_sql(sql_template, context=dict(
            base_relation=dbt_adapter.create_relation(model, base=True),
            curr_relation=dbt_adapter.create_relation(model, base=False),
            primary_key=primary_key,
            columns=columns,
            limit=1000,
        ))

        _, table = dbt_adapter.execute(sql, fetch=True)
        self.check_cancel()

        return DataFrame.from_agate(table)

    def execute(self):

        from recce.adapter.dbt_adapter import DbtAdapter
        dbt_adapter: DbtAdapter = default_context().adapter

        with dbt_adapter.connection_named("value diff"):
            self.connection = dbt_adapter.get_thread_connection()

            primary_key: Union[str, List[str]] = self.params.primary_key
            model: str = self.params.model
            columns: List[str] = self.params.columns

            self._verify_dbt_packages_deps(dbt_adapter)
            self.check_cancel()

            self._verify_primary_key(dbt_adapter, primary_key, model)
            self.check_cancel()

            return self._query_value_diff(dbt_adapter, primary_key, model, columns)

    def cancel(self):
        from recce.adapter.dbt_adapter import DbtAdapter
        if self.connection:
            adapter: DbtAdapter = default_context().adapter
            with adapter.connection_named("cancel"):
                adapter.cancel(self.connection)


class ValueDiffDetailTaskResultDiffer(TaskResultDiffer):

    def _check_result_changed_fn(self, result):
        diff_data = result.get('data')
        if diff_data is None or len(diff_data) == 0:
            return None

        # TODO: Implement detailed information of values changed
        return dict(values_changed={})


class ValueDiffCheckValidator(CheckValidator):

    def validate_check(self, check: Check):
        try:
            ValueDiffParams(**check.params)
        except Exception as e:
            raise ValueError(f"Invalid check: {str(e)}")



import pytest


def test_default_validator():
    from recce.tasks.core import CheckValidator

    CheckValidator().validate({
        "name": "test",
        "type": "row_count_diff",
        "params": {},
    })

    with pytest.raises(ValueError):
        CheckValidator().validate({
            "type": "row_count_diff",
        })


def test_query_diff_validator():
    from recce.tasks.query import QueryDiffCheckValidator
    QueryDiffCheckValidator().validate({
        "name": "test",
        "type": "query_diff",
        "params": {
            "sql_template": "select * from {{ model }}"
        },
    })
    with pytest.raises(ValueError):
        QueryDiffCheckValidator().validate({
            "name": "test",
            "type": "query_diff",
        })


def test_lineage_diff_validator():
    from recce.tasks.lineage import LineageDiffCheckValidator
    LineageDiffCheckValidator().validate({
        "name": "test",
        "type": "lineage_diff",
        "params": {},
    })
    LineageDiffCheckValidator().validate({
        "name": "test",
        "type": "lineage_diff",
        "view_options": {},
    })
    with pytest.raises(ValueError):
        LineageDiffCheckValidator().validate({
            "type": "lineage_diff",
        })



import typing
from typing import Optional, Tuple, List

from pydantic import BaseModel

from .core import Task, TaskResultDiffer, CheckValidator
from .dataframe import DataFrame
from .valuediff import ValueDiffMixin
from ..core import default_context
from ..exceptions import RecceException
from ..models import Check

QUERY_LIMIT = 2000

if typing.TYPE_CHECKING:
    import agate


class QueryMixin:
    @classmethod
    def execute_sql_with_limit(
        cls,
        sql_template,
        base: bool = False,
        limit: Optional[int] = None
    ) -> Tuple['agate.Table', bool]:
        """
        Execute a SQL template and return the result as an agate table.
        :param sql_template: SQL template to execute
        :param base: Whether to run the query on the base environment
        :param limit: Limit the number of rows returned
        :return: Tuple of agate table and whether there are more rows to fetch
        """
        from jinja2.exceptions import TemplateSyntaxError
        dbt_adapter = default_context().adapter
        try:
            sql = dbt_adapter.generate_sql(sql_template, base)

            if limit is None:
                _, result = dbt_adapter.execute(sql, fetch=True, auto_begin=True)
                return result, False
            else:
                _, result = dbt_adapter.execute(sql, fetch=True, auto_begin=True, limit=limit + 1)
                if len(result.rows) > limit:
                    return result.limit(limit), True
                return result, False

        except TemplateSyntaxError as e:
            raise RecceException(f"Jinja template error: line {e.lineno}: {str(e)}")

    @classmethod
    def execute_sql(cls, sql_template, base: bool = False) -> 'agate.Table':
        result, _ = cls.execute_sql_with_limit(sql_template, base)
        return result

    @staticmethod
    def close_connection(connection):
        dbt_adapter = default_context().adapter
        with dbt_adapter.connection_named("cancel query"):
            dbt_adapter.cancel(connection)


class QueryParams(BaseModel):
    sql_template: str


class QueryResult(DataFrame):
    pass


class QueryDiffParams(BaseModel):
    sql_template: str
    base_sql_template: Optional[str] = None
    primary_keys: Optional[List[str]] = None
    current_model: Optional[str] = None


class QueryTask(Task, QueryMixin):
    is_base = False

    def __init__(self, params: dict):
        super().__init__()
        self.params = QueryParams(**params)
        self.connection = None

    def execute_dbt(self):
        from recce.adapter.dbt_adapter import DbtAdapter
        dbt_adapter: DbtAdapter = default_context().adapter

        limit = QUERY_LIMIT
        with dbt_adapter.connection_named("query"):
            self.connection = dbt_adapter.get_thread_connection()

            sql_template = self.params.sql_template
            table, more = self.execute_sql_with_limit(sql_template, base=self.is_base, limit=limit)
            self.check_cancel()

            return DataFrame.from_agate(table, limit=limit, more=more)

    def execute_sqlmesh(self):
        from ..adapter.sqlmesh_adapter import SqlmeshAdapter
        sqlmesh_adapter: SqlmeshAdapter = default_context().adapter

        sql = self.params.get('sql_template')
        limit = QUERY_LIMIT
        df, more = sqlmesh_adapter.fetchdf_with_limit(sql, base=self.is_base, limit=limit)
        return DataFrame.from_pandas(df, limit=limit, more=more)

    def execute(self):
        context = default_context()

        if context.adapter_type == 'sqlmesh':
            return self.execute_sqlmesh()
        else:
            return self.execute_dbt()

    def cancel(self):
        super().cancel()
        if self.connection:
            self.close_connection(self.connection)


class QueryBaseTask(QueryTask):
    is_base = True


class QueryDiffResult(BaseModel):
    base: Optional[DataFrame] = None
    current: Optional[DataFrame] = None
    diff: Optional[DataFrame] = None


class QueryDiffTask(Task, QueryMixin, ValueDiffMixin):
    def __init__(self, params):
        super().__init__()
        self.params = QueryDiffParams(**params)
        self.connection = None
        self.legacy_surrogate_key = True

    def _query_diff(self, dbt_adapter, sql_template: str, base_sql_template: Optional[str] = None,
                    preview_change: bool = False):
        limit = QUERY_LIMIT

        self.connection = dbt_adapter.get_thread_connection()
        if preview_change:
            base, base_more = self.execute_sql_with_limit(base_sql_template, base=False, limit=limit)
        else:
            base, base_more = self.execute_sql_with_limit(base_sql_template or sql_template, base=True, limit=limit)
        self.check_cancel()

        current, current_more = self.execute_sql_with_limit(sql_template, base=False, limit=limit)
        self.check_cancel()

        return QueryDiffResult(
            base=DataFrame.from_agate(base, limit=limit, more=base_more),
            current=DataFrame.from_agate(current, limit=limit, more=current_more)
        )

    def _query_diff_join(self, dbt_adapter, sql_template: str, primary_keys: List[str],
                         base_sql_template: Optional[str] = None, preview_change: bool = False):

        query_template = r"""
            {% set a_query %}
                {{ base_query }}
            {% endset %}

            {% set b_query %}
                {{ current_query }}
            {% endset %}

            {{ audit_helper.compare_queries(
                a_query=a_query,
                b_query=b_query,
                primary_key=__PRIMARY_KEY__,
                summarize=False,
            ) }} limit {{ limit }}
            """

        if len(primary_keys) > 1:
            self._verify_dbt_packages_deps(dbt_adapter)
            self.check_cancel()

            if self.legacy_surrogate_key:
                new_primary_key = 'dbt_utils.surrogate_key(primary_key)'
            else:
                new_primary_key = 'dbt_utils.generate_surrogate_key(primary_key)'
        else:
            new_primary_key = 'primary_key'
        query_template = query_template.replace('__PRIMARY_KEY__', new_primary_key)

        if preview_change:
            base_query = dbt_adapter.generate_sql(base_sql_template, base=False)
        else:
            base_query = dbt_adapter.generate_sql(base_sql_template or sql_template, base=True)
        current_query = dbt_adapter.generate_sql(sql_template, base=False)

        sql = dbt_adapter.generate_sql(query_template, context=dict(
            base_query=base_query,
            current_query=current_query,
            primary_key=primary_keys if len(primary_keys) != 1 else primary_keys[0],
            limit=QUERY_LIMIT,
        ))

        _, table = dbt_adapter.execute(sql, fetch=True)
        self.check_cancel()

        return QueryDiffResult(
            diff=DataFrame.from_agate(table)
        )

    @staticmethod
    def _select_single_model(model_name):
        return f'select * from {{{{ ref("{model_name}") }}}}'

    def execute_dbt(self):
        from recce.adapter.dbt_adapter import DbtAdapter
        dbt_adapter: DbtAdapter = default_context().adapter

        with dbt_adapter.connection_named("query"):
            preview_change = False
            sql_template = self.params.sql_template
            primary_keys = self.params.primary_keys
            base_sql_template = self.params.base_sql_template
            if self.params.current_model:
                base_sql_template = self._select_single_model(self.params.current_model)
                preview_change = True

            if primary_keys:
                return self._query_diff_join(dbt_adapter, sql_template, primary_keys,
                                             base_sql_template=base_sql_template, preview_change=preview_change)

            return self._query_diff(dbt_adapter, sql_template, base_sql_template=base_sql_template,
                                    preview_change=preview_change)

    def _sqlmesh_query_diff(self, sql, base_sql=None):
        from ..adapter.sqlmesh_adapter import SqlmeshAdapter

        sqlmesh_adapter: SqlmeshAdapter = default_context().adapter

        limit = QUERY_LIMIT
        base, base_more = sqlmesh_adapter.fetchdf_with_limit(base_sql or sql, base=True, limit=limit)
        curr, curr_more = sqlmesh_adapter.fetchdf_with_limit(sql, base=False, limit=limit)
        return QueryDiffResult(
            base=DataFrame.from_pandas(base, limit=limit, more=base_more),
            current=DataFrame.from_pandas(curr, limit=limit, more=curr_more)
        )

    def _sqlmesh_query_diff_join(self, sql, primary_keys, base_sql=None):
        from ..adapter.sqlmesh_adapter import SqlmeshAdapter

        sqlmesh_adapter: SqlmeshAdapter = default_context().adapter

        limit = QUERY_LIMIT
        expr_base = sqlmesh_adapter.replace_virtual_tables(base_sql or sql, base=True)
        expr_curr = sqlmesh_adapter.replace_virtual_tables(sql, base=False)
        import sqlglot as g

        expr = g.select(
            '*',
        ).with_(
            'a', as_=expr_base
        ).with_(
            'b', as_=expr_curr
        ).with_(
            'a_interset_b', as_='select * from a intersect select * from b'
        ).with_(
            'a_except_b', as_='select * from a except select * from b'
        ).with_(
            'b_except_a', as_='select * from b except select * from a'
        ).with_(
            'all_records',
            as_='''
            SELECT
              *,
              TRUE AS in_a,
              TRUE AS in_b
            FROM a_interset_b
            UNION ALL
            SELECT
              *,
              TRUE AS in_a,
              FALSE AS in_b
            FROM a_except_b
            UNION ALL
            SELECT
              *,
              FALSE AS in_a,
              TRUE AS in_b
            FROM b_except_a
            '''
        ).with_(
            'final',
            as_=f'''
                    select * from all_records
                    where not (in_a and in_b)
                    order by {", ".join(primary_keys)}, in_a desc, in_b desc
                    '''
        ).from_('final').limit(1000)
        diff, diff_more = sqlmesh_adapter.fetchdf_with_limit(expr, limit=limit)
        return QueryDiffResult(
            diff=DataFrame.from_pandas(diff, limit=limit, more=diff_more)
        )

    def execute_sqlmesh(self):
        sql = self.params.sql_template
        primary_keys = self.params.primary_keys
        base_sql = self.params.base_sql_template

        if primary_keys:
            return self._sqlmesh_query_diff_join(sql, primary_keys, base_sql=base_sql)
        else:
            return self._sqlmesh_query_diff(sql, base_sql=base_sql)

    def execute(self):
        context = default_context()

        if context.adapter_type == 'sqlmesh':
            return self.execute_sqlmesh()
        else:
            return self.execute_dbt()

    def cancel(self):
        super().cancel()
        if self.connection:
            self.close_connection(self.connection)


class QueryDiffResultDiffer(TaskResultDiffer):
    def _check_result_changed_fn(self, result):
        base = result.get('base')
        current = result.get('current')
        diff = result.get('diff')

        if diff is None:
            return TaskResultDiffer.diff(base, current)
        else:
            diff_data = diff.get('data')
            if diff_data is None or len(diff_data) == 0:
                return None

            # TODO: Implement detailed information of values changed
            return dict(values_changed={})


class QueryCheckValidator(CheckValidator):
    def validate_check(self, check: Check):
        try:
            QueryParams(**check.params)
        except Exception as e:
            raise ValueError(f"Invalid check: {str(e)}")


class QueryDiffCheckValidator(CheckValidator):

    def validate_check(self, check: Check):
        try:
            QueryDiffParams(**check.params)
        except Exception as e:
            raise ValueError(f"Invalid check: {str(e)}")



import pytest


def test_validator():
    from recce.tasks.lineage import LineageDiffCheckValidator
    validator = LineageDiffCheckValidator()

    def validate(params: dict):
        validator.validate({
            'name': 'test',
            'type': 'schema_diff',
            'params': params,
        })

    # Select all models
    validate({})

    # Select by selector
    validate({
        'select': 'customers',
        'exclude': 'customers',
        'packages': ['jaffle_shop'],
        'view_mode': 'all',
    })

    # packages should be an array
    with pytest.raises(ValueError):
        validate({
            'packages': 'jaffle_shop',
        })

    # view_mode should be 'all' or 'changed_models'
    validate({
        'view_mode': None,
    })
    validate({
        'view_mode': 'all',
    })
    with pytest.raises(ValueError):
        validate({
            'view_mode': 'abc',
        })



#!/usr/bin/env python
import os
from distutils.core import setup

from setuptools import find_packages  # type: ignore


def _get_version():
    version_file = os.path.normpath(os.path.join(os.path.dirname(__file__), 'recce', 'VERSION'))
    with open(version_file) as fh:
        version = fh.read().strip()
        return version


setup(name='recce',
      version=_get_version(),
      description='Environment diff tool for dbt',
      long_description=open('README.md').read(),
      long_description_content_type='text/markdown',
      author='InfuseAI Dev Team',
      author_email='dev@infuseai.io',
      url='https://github.com/InfuseAI/recce',
      entry_points={
          'console_scripts': ['recce = recce.cli:cli']
      },
      python_requires=">=3.9",
      packages=find_packages(),
      install_requires=[
          'boto3',
          'requests>=2.28.1',
          'ruamel.yaml>=0.18.6',
          'click>=7.1',
          'deepdiff>=7.0,<8.0',
          'portalocker',
          'fastapi',
          'itsdangerous',
          'uvicorn',
          'pydantic',
          'jinja2',
          'requests>=2.28.1',
          'rich>=12.0.0',
          'sentry-sdk',
          'watchdog',
          'websockets',
          'py-markdown-table',
          'python-dateutil',
          'python-multipart',
          'GitPython',
          'PyGithub',
          'sqlglot',
      ],
      tests_require=['pytest'],
      extras_require={
          'dev': [
              'pytest>=4.6',
              'pytest-flake8',
              'flake8==3.9.2',
              'pytest-mypy',
              'pytest-cov',
              'twine',
              'tox',
              'pandas',
              'httpx',
          ],
      },
      project_urls={
          "Bug Tracker": "https://github.com/InfuseAI/recce/issues",
      },
      classifiers=[
          "Programming Language :: Python :: 3.9",
          "Programming Language :: Python :: 3.10",
          "Programming Language :: Python :: 3.11",
          "License :: OSI Approved :: Apache Software License",
          "Operating System :: OS Independent",
          "Development Status :: 4 - Beta"
      ],
      package_data={"recce": ['VERSION', 'data/**', 'event/SENTRY_DNS', 'event/CONFIG']})



import asyncio
import json
import logging
import os
import uuid
from contextlib import asynccontextmanager
from dataclasses import dataclass
from pathlib import Path
from typing import Optional, Any, Set, Annotated, Literal

from fastapi import FastAPI, HTTPException, Request, WebSocket, UploadFile, Response, BackgroundTasks, Form
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import PlainTextResponse
from fastapi.staticfiles import StaticFiles
from pydantic import ValidationError, BaseModel
from starlette.middleware.gzip import GZipMiddleware
from starlette.middleware.sessions import SessionMiddleware
from starlette.websockets import WebSocketDisconnect

from . import __version__, event
from .apis.check_api import check_router
from .apis.run_api import run_router
from .config import RecceConfig
from .core import load_context, default_context
from .event import log_api_event, log_single_env_event
from .exceptions import RecceException
from .run import load_preset_checks
from .state import RecceStateLoader

logger = logging.getLogger('uvicorn')


@dataclass
class AppState:
    state_loader: Optional[RecceStateLoader] = None
    kwargs: Optional[dict] = None
    flag: Optional[dict] = None


@asynccontextmanager
async def lifespan(fastapi: FastAPI):
    from .core import load_context
    from rich.console import Console

    console = Console()
    app_state: AppState = app.state
    state_loader = app_state.state_loader
    kwargs = app_state.kwargs
    ctx = load_context(**kwargs, state_loader=state_loader)
    ctx.start_monitor_artifacts(callback=dbt_artifacts_updated_callback)
    if app_state.flag.get("single_env_onboarding", False) is True:
        # [Experiment 2] Start with Single Environment
        ctx.start_monitor_base_env(callback=dbt_env_updated_callback)
        log_single_env_event()

    # Initialize Recce Config
    config = RecceConfig(config_file=kwargs.get('config'))
    if state_loader.state is None:
        preset_checks = config.get('checks', [])
        if preset_checks and len(preset_checks) > 0:
            console.rule("Loading Preset Checks")
            load_preset_checks(preset_checks)

    from recce.event import log_load_state
    log_load_state(command='server')

    yield

    state_loader.export(ctx.export_state())

    ctx.stop_monitor_artifacts()
    if app_state.flag.get("single_env_onboarding", False):
        ctx.stop_monitor_base_env()


app = FastAPI(lifespan=lifespan)


def verify_json_file(file_path: str) -> bool:
    try:
        with open(file_path, 'r') as f:
            json.load(f)
    except Exception:
        return False
    return True


def dbt_artifacts_updated_callback(file_changed_event: Any):
    src_path = Path(file_changed_event.src_path)
    target_type = src_path.parent.name
    file_name = src_path.name

    if not verify_json_file(file_changed_event.src_path):
        logger.debug('Skip to refresh the artifacts because the file is not updated completely.')
        return

    logger.info(
        f'Detect {target_type} file {file_changed_event.event_type}: {file_name}')
    ctx = load_context()
    ctx.refresh_manifest(file_changed_event.src_path)
    broadcast_command = {
        'command': 'refresh',
        'event': {
            'eventType': file_changed_event.event_type,
            'srcPath': file_changed_event.src_path
        }
    }
    payload = json.dumps(broadcast_command)
    asyncio.run(broadcast(payload))


def dbt_env_updated_callback():
    logger.info("Detect 'manifest.json' and 'catalog.json' are generated under 'target-base' directory")
    broadcast_command = {
        'command': 'relaunch',
    }
    payload = json.dumps(broadcast_command)
    asyncio.run(broadcast(payload))


clients = set()

origins = [
    "http://localhost:3000",
    "http://localhost:3001",
]

app.add_middleware(GZipMiddleware, minimum_size=1000)
app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)
app.add_middleware(
    SessionMiddleware,
    secret_key=uuid.uuid4(),
)


@app.middleware("http")
async def set_context_by_cookie(request: Request, call_next):
    response = await call_next(request)

    user_id_in_cookie = request.cookies.get('recce_user_id')
    user_id = event.get_user_id()

    if event.is_anonymous_tracking() is False:
        # Disable anonymous tracking
        user_id = None

    if user_id_in_cookie is None or user_id_in_cookie != user_id:
        response.set_cookie(key='recce_user_id', value=user_id)
    return response


@app.middleware("http")
async def disable_cache(request: Request, call_next):
    response = await call_next(request)

    # disable cache for '/' and '/index.html'
    if request.url.path in ['/', '/index.html']:
        response.headers['Cache-Control'] = 'no-store'

    return response


@app.get("/api/health")
async def health_check(request: Request):
    return {"status": "ok"}


@app.get("/api/flag")
async def config_flag():
    app_state: AppState = app.state
    flag = app_state.flag
    return flag


@app.post("/api/onboarding/completed", status_code=204)
async def mark_onboarding_completed():
    context = default_context()
    context.mark_onboarding_completed()
    app.state.flag['show_onboarding_guide'] = False


@app.post("/api/relaunch-hint/completed", status_code=204)
async def mark_relaunch_hint_completed():
    app.state.flag['show_relaunch_hint'] = False


@app.get("/api/info")
async def get_info():
    """
    Get the information of the current context.
    """
    context = default_context()
    demo = os.environ.get('DEMO', False)

    if demo:
        state = context.export_demo_state()
    else:
        state = context.export_state()

    support_tasks = context.support_tasks()
    if context.state_loader and context.state_loader.state_file:
        filename = os.path.basename(context.state_loader.state_file)
    else:
        filename = None

    lineage_diff = context.get_lineage_diff()

    try:
        info = {
            'adapter_type': context.adapter_type,
            'review_mode': context.review_mode,
            'git': state.git.to_dict() if state.git else None,
            'pull_request': state.pull_request.to_dict() if state.pull_request else None,
            'lineage': lineage_diff,
            'demo': bool(demo),
            'cloud_mode': context.state_loader.cloud_mode,
            'file_mode': context.state_loader.state_file is not None,
            'filename': filename,
            'support_tasks': support_tasks,
        }

        if context.adapter_type == 'sqlmesh':
            from recce.adapter.sqlmesh_adapter import SqlmeshAdapter
            sqlmesh_adapter: SqlmeshAdapter = context.adapter
            info['sqlmesh'] = {
                'base_env': sqlmesh_adapter.base_env.name,
                'current_env': sqlmesh_adapter.curr_env.name,
            }

        return info
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e))


class SelectNodesInput(BaseModel):
    select: Optional[str] = None
    exclude: Optional[str] = None
    packages: Optional[list[str]] = None
    view_mode: Optional[Literal['all', 'changed_models']] = None


class SelectNodesOutput(BaseModel):
    nodes: Set[str] = []


@app.post("/api/select", response_model=SelectNodesOutput)
async def select_nodes(input: SelectNodesInput):
    context = default_context()

    if context.adapter_type != 'dbt':
        raise HTTPException(status_code=400, detail='Only dbt adapter is supported')

    try:
        nodes = context.adapter.select_nodes(
            select=input.select,
            exclude=input.exclude,
            packages=input.packages,
            view_mode=input.view_mode,
        )
        nodes = [node for node in nodes if not node.startswith('test.')]
        return SelectNodesOutput(nodes=nodes)
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e))


@app.get("/api/model/{model_id}")
async def get_columns(model_id: str):
    context = default_context()
    try:
        return {
            'model': {
                'base': context.get_model(model_id, base=True),
                'current': context.get_model(model_id, base=False)
            }
        }
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e))


@app.post("/api/save", response_class=PlainTextResponse, status_code=200)
async def save_handler():
    """
    Save the in-memory state to the state file
    """
    try:
        # Sync the state file
        context = default_context()
        log_api_event('save', dict(state_loader_mode=context.state_loader_mode()))
        state_loader = context.state_loader
        if not state_loader.cloud_mode and state_loader.state_file is None:
            raise RecceException('Not file mode or cloud mode')

        context.sync_state('overwrite')
    except RecceException as e:
        raise HTTPException(status_code=400, detail=e.message)


class SaveAsOrRenameInput(BaseModel):
    # The filename. The filename should not contain directory.
    filename: Optional[str] = None
    # Overwrite the file if it exists
    overwrite: Optional[bool] = False


def saveas_or_rename(input: SaveAsOrRenameInput, rename: bool = False):
    context = default_context()
    state_loader = context.state_loader
    if state_loader.cloud_mode:
        raise RecceException('Cloud mode does not support rename')

    new_filename = input.filename
    if os.path.dirname(new_filename):
        raise RecceException('The new filename should not contain directory')
    if not new_filename.endswith('.json'):
        raise RecceException('The new filename should end with .json')

    old_path = state_loader.state_file
    if old_path:
        old_dir = os.path.dirname(state_loader.state_file)
        old_filename = os.path.basename(state_loader.state_file)
        if old_filename == new_filename:
            raise RecceException('The new filename is the same as the current filename')
        new_path = os.path.join(old_dir, new_filename)
    else:
        new_path = new_filename

    if os.path.exists(new_path):
        if os.path.isdir(new_path):
            raise HTTPException(status_code=400, detail=f'The file {new_path} exists and is a directory')

        if not input.overwrite:
            raise HTTPException(status_code=409, detail=f'The file {new_filename} already exists')

    state_loader.state_file = new_path
    context.sync_state('overwrite')
    if rename and os.path.exists(old_path):
        os.remove(old_path)


@app.post("/api/save-as", response_class=PlainTextResponse, status_code=200)
async def save_as_handler(input: SaveAsOrRenameInput):
    """
    Save the state to a new file
    """
    context = default_context()
    try:
        log_api_event('saveas', dict(state_loader_mode=context.state_loader_mode()))
        saveas_or_rename(input, rename=False)
    except RecceException as e:
        raise HTTPException(status_code=400, detail=e.message)


@app.post("/api/rename", response_class=PlainTextResponse, status_code=200)
async def rename_handler(input: SaveAsOrRenameInput):
    """
    Rename the state to a new file
    """
    context = default_context()
    try:
        log_api_event('rename', dict(state_loader_mode=context.state_loader_mode()))
        saveas_or_rename(input, rename=True)
    except RecceException as e:
        raise HTTPException(status_code=400, detail=e.message)


@app.post("/api/export", response_class=PlainTextResponse, status_code=200)
async def export_handler():
    """
    Export the recce state to the client.
    """
    context = default_context()
    try:
        log_api_event('export', dict(state_loader_mode=context.state_loader_mode()))
        return context.export_state().to_json()
    except RecceException as e:
        raise HTTPException(status_code=400, detail=e.message)


@app.post("/api/import", status_code=200)
async def import_handler(
    file: Annotated[UploadFile, Form()],
    checks_only: Annotated[bool, Form()],
    background_tasks: BackgroundTasks
):
    """
    Import the recce state from the client.
    """
    from recce.state import RecceState
    context = default_context()
    try:
        log_api_event('import', dict(state_loader_mode=context.state_loader_mode()))
        content = await file.read()
        state = RecceState.from_json(content)

        if checks_only:
            import_checks = context.import_checks(state)
            background_tasks.add_task(context.sync_state, "overwrite")
            return {"runs": 0, "checks": import_checks}

        import_runs, import_checks = context.import_state(state)
        return {"runs": import_runs, "checks": import_checks}
    except ValidationError as e:
        raise HTTPException(status_code=400, detail=str(e))
    except RecceException as e:
        raise HTTPException(status_code=400, detail=e.message)


class SyncStateInput(BaseModel):
    method: Optional[str] = None


@app.post("/api/sync", status_code=202)
async def sync_handler(input: SyncStateInput, response: Response, background_tasks: BackgroundTasks):
    """
    Sync the state with the external storage. (two-way sync)

    This is used to sync the state with the external (local or cloud) storage. There are three methods:
    - overwrite: Overwrite the external storage with the in-memory state.
    - revert: Revert the in-memory state with the external storage.
    - merge: Merge the state between the in-memory and external storage.
    """
    context = default_context()
    state_loader = context.state_loader
    method = input.method
    log_api_event('sync', dict(
        state_loader_mode=context.state_loader_mode(),
        method=method,
    ))

    if not method:
        is_conflict = state_loader.check_conflict()
        if is_conflict:
            raise HTTPException(status_code=409, detail='Conflict detected')
        method = 'overwrite'

    is_syncing = state_loader.state_lock.locked()
    if is_syncing:
        response.status_code = 208
        return {"status": "syncing"}

    def reload_state():
        ctx = default_context()
        ctx.sync_state(method)

    background_tasks.add_task(reload_state)
    response.status_code = 202
    return {"status": "request accepted"}


@app.get("/api/sync", status_code=200)
async def sync_status(response: Response):
    """
    Get the sync status.
    """
    context = default_context()
    if context.state_loader.state_lock.locked():
        response.status_code = 208
        return {"status": "syncing"}

    response.status_code = 200
    return {"status": "idle"}


@app.get("/api/version")
async def version():
    try:
        return __version__
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e))


@app.websocket("/api/ws")
async def websocket_endpoint(websocket: WebSocket):
    await websocket.accept()
    clients.add(websocket)
    try:
        while True:
            data = await websocket.receive_text()
            if data == 'ping':
                await websocket.send_text('pong')
    except WebSocketDisconnect:
        clients.remove(websocket)


async def broadcast(data: str):
    for client in clients:
        await client.send_text(data)


api_prefix = '/api'
app.include_router(check_router, prefix=api_prefix)
app.include_router(run_router, prefix=api_prefix)

static_folder_path = Path(__file__).parent / 'data'
app.mount("/", StaticFiles(directory=static_folder_path, html=True), name="static")



import asyncio
from typing import Optional, List
from uuid import UUID

from fastapi import APIRouter, HTTPException, Query
from pydantic import BaseModel

from recce.apis.run_func import submit_run, cancel_run, materialize_run_results
from recce.event import log_api_event
from recce.exceptions import RecceException
from recce.models import RunDAO

run_router = APIRouter(tags=['run'])


class CreateRunIn(BaseModel):
    type: str
    params: dict
    check_id: Optional[str] = None
    nowait: Optional[bool] = False
    track_props: Optional[dict] = None


@run_router.post("/runs", status_code=201)
async def create_run_handler(input: CreateRunIn):
    log_api_event('create_run', dict(
        type=input.type,
        track_props=input.track_props,
    ))
    try:
        run, future = submit_run(input.type, input.params)
    except RecceException as e:
        raise HTTPException(status_code=400, detail=str(e))

    if input.nowait:
        return run
    else:
        run.result = await future
        return run


@run_router.post("/runs/{run_id}/cancel")
async def cancel_run_handler(run_id: UUID):
    try:
        cancel_run(run_id)
    except NotImplementedError:
        pass


@run_router.get("/runs/{run_id}/wait")
async def wait_run_handler(run_id: UUID, timeout: int = Query(None, description="Maximum number of seconds to wait")):
    run = RunDAO().find_run_by_id(run_id)
    if run is None:
        raise HTTPException(status_code=404, detail='Not Found')

    start_time = asyncio.get_event_loop().time()
    while run.result is None and run.error is None:
        await asyncio.sleep(1)
        if timeout is not None and (asyncio.get_event_loop().time() - start_time) > timeout:
            break
    return run


@run_router.get("/runs", status_code=200)
async def list_run_handler():
    runs = RunDAO().list() or []

    result = [{
        'run_id': run.run_id,
        'run_at': run.run_at,
        'name': run.name,
        'type': run.type,
        'params': run.params,
        'status': run.status,
        'check_id': run.check_id,
    } for run in runs]

    # sort by run_at
    result = sorted(result, key=lambda x: x['run_at'], reverse=True)

    return result


class SearchRunsIn(BaseModel):
    type: str
    params: dict
    limit: Optional[int] = None


@run_router.post("/runs/search", status_code=200)
async def search_runs_handler(search: SearchRunsIn):
    runs = RunDAO().list()

    result = []
    for run in runs:
        if run.type.value != search.type:
            continue
        if not all(search.params[key] == run.params.get(key) for key in search.params.keys()):
            continue

        result.append(run)

    if search.limit:
        return result[-search.limit:]

    return result


class AggregateRunsIn(BaseModel):
    class AggregateFilter(BaseModel):
        nodes: Optional[List[str]] = None

    filter: Optional[AggregateFilter] = None


@run_router.post("/runs/aggregate", status_code=200)
async def aggregate_runs_handler(input: AggregateRunsIn):
    try:
        runs = RunDAO().list()
        nodes = input.filter.nodes if input.filter and input.filter.nodes else None
        result = materialize_run_results(runs, nodes=nodes)
        return result
    except Exception as e:
        raise HTTPException(status_code=405, detail=str(e))
